/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2005 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <config.h>
#include <xpl.h>
#include <logger.h>
#include <mdb.h>
#include <nmap.h>
#include <nmlib.h>

#include "nmapdp.h"

typedef struct _NmapDStoreInfo {
    NmapDocumentTypes type;

    unsigned int length;
    unsigned int indexed;

    struct {
        unsigned int store;
        unsigned int doc;
    } start;

    struct {
        unsigned char source[22];
        unsigned char store[22];
    } version;

    unsigned char guid[NMAP_GUID_LENGTH + 1];
    unsigned char resource[XPL_MAX_PATH + 1];
} NmapDStoreInfo;

static int 
DStoreTranBegin(NMAPClient *client, BOOL exclusive)
{
    int ccode;
    int count = 30;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt;

    if (((ccode = NmapSqlPrepare(handle, NMAP_SQLTRAN_BEGIN, &(client->dstore.sql.begin))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_SQLTRAN_BEGIN_SHARED, &(client->dstore.sql.shared))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_SQLTRAN_END, &(client->dstore.sql.end))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_SQLTRAN_ROLLBACK, &(client->dstore.sql.abort))) == SQLITE_OK)) {
        if (exclusive) {
            stmt = client->dstore.sql.begin;
        } else {
            stmt = client->dstore.sql.shared;
        }

        do {
            ccode = sqlite3_step(stmt);

            sqlite3_reset(stmt);
            if (ccode == SQLITE_DONE) {
                client->states |= NMAP_CLIENT_DSTORE_TRANSACTION;

                return(SQLITE_DONE);
            }

            if (ccode == SQLITE_BUSY) {
                XplDelay(333);

                count--;
                continue;
            }
        } while ((ccode == SQLITE_BUSY) && count);

#if defined(DEBUG)
        XplConsolePrintf("NMAP [%d]: Begin document transaction for \"%s\" failed; %s\r\n", __LINE__, client->user, sqlite3_errmsg(handle));
#endif
    }

    return(ccode);
}

/*
    The sqlite statement preparation is performed by DocStoreTranBegin which is 
    therefore a pre-requisite for usage.
*/
static int 
DStoreTranEnd(NMAPClient *client)
{
    register int ccode;

    if (client->states & NMAP_CLIENT_DSTORE_TRANSACTION) {
        client->states &= ~NMAP_CLIENT_DSTORE_TRANSACTION;

        ccode = sqlite3_step(client->dstore.sql.end);

        sqlite3_reset(client->dstore.sql.end);

#if defined(DEBUG)
        if (ccode != SQLITE_DONE) {
            XplConsolePrintf("NMAP [%d]: End document transaction for \"%s\" failed; %s\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }
#endif

        return(ccode);
    }

    return(SQLITE_DONE);
}

/*
    The sqlite statement preparation is performed by DocStoreTranBegin which is 
    therefore a pre-requisite for usage.
*/
static void 
DStoreTranAbort(NMAPClient *client)
{
    register int ccode;

    if (client->states & NMAP_CLIENT_DSTORE_TRANSACTION) {
        client->states &= ~NMAP_CLIENT_DSTORE_TRANSACTION;

        ccode = sqlite3_step(client->dstore.sql.abort);

        sqlite3_reset(client->dstore.sql.abort);

#if defined(DEBUG)
        if (ccode != SQLITE_DONE) {
            XplConsolePrintf("NMAP [%d]: Abort document transaction for \"%s\" failed; %s\r\n", __LINE__, client->mail.name, sqlite3_errmsg(client->dstore.handle));
        }
#endif

        ccode = sqlite3_step(client->dstore.sql.end);

        sqlite3_reset(client->dstore.sql.end);

#if defined(DEBUG)
        if (ccode != SQLITE_DONE) {
            XplConsolePrintf("NMAP [%d]: End document transaction for \"%s\" failed; %s\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }
#endif
    }

    return;
}

static int 
UpgradeNmapMailStore(NMAPClient *client)
{
    int ccode;
    unsigned char line[XPL_MAX_PATH + 24];
    unsigned char guid[XPL_MAX_PATH + 1];
    NLockStruct *lock = client->mail.lock;

    client->mail.lock = NULL;
    GuidAlloc(guid);
    guid[NMAP_GUID_LENGTH] = '\0';

    if (ReadNLockAquire(&client->mail.lock, &client->userHash, client->user, client->mail.name, 3000) == NLOCK_AQUIRED) {
        ccode = ParseMailStore(client);

        client->mail.info.used = 0;
        client->mail.info.count = 0;

        if (ccode == NMAP_OK) {
            sprintf(client->path, "%s/%s/nmap.db", client->store, client->user);
            ccode = sqlite3_open(client->path, &client->dstore.handle);
        } else {
           ccode = SQLITE_ERROR;
        }

        if (ccode == SQLITE_OK) {
            sprintf(line, "INSERT INTO map (guid, type, start, toDoc, length, resource, indexed, modified) VALUES ('%s', %d, 0, 0, 0, '%s', 0, julianday(DATETIME('now')));", guid, NMAP_DOCTYPE_MAIL_FOLDER, client->mail.name);
            ccode = sqlite3_exec(client->dstore.handle, line, NULL, NULL, NULL);
        }

        while (ccode == SQLITE_OK) {
            sprintf(line, "ATTACH \"%s/%s/%s"NMAP_MAIL_INDEX_EXT"\" AS mail;", client->store, client->user, client->mail.name);
            ccode = sqlite3_exec(client->dstore.handle, line, NULL, NULL, NULL);
            if (ccode == SQLITE_OK) {
                ccode = DStoreTranBegin(client, TRUE);
            } else {
                XplConsolePrintf("NMAP [%d]: Upgrade mail folder \"%s\" for \"%s\"; attach error \"%s\".\r\n", __LINE__, client->mail.name, client->user, sqlite3_errmsg(client->dstore.handle));
                break;
            }

            if (ccode == SQLITE_DONE) {
                ccode = sqlite3_exec(client->dstore.handle, "INSERT INTO map (guid, start, toDoc, length, modified) SELECT guid, start, toAuth, msgSize, modified FROM mail.message;", NULL, NULL, NULL);
            } else {
                sqlite3_exec(client->dstore.handle, "DETACH mail;", NULL, NULL, NULL);
                break;
            }

            if (ccode == SQLITE_OK) {
                sprintf(line, "UPDATE map SET resource='%s', type=%d WHERE resource ISNULL;", client->mail.name, NMAP_DOCTYPE_MAIL);
                ccode = sqlite3_exec(client->dstore.handle, line, NULL, NULL, NULL);
            } else {
                XplConsolePrintf("NMAP [%d]: Upgrade mail folder \"%s\" for \"%s\"; insert error \"%s\".\r\n", __LINE__, client->mail.name, client->user, sqlite3_errmsg(client->dstore.handle));

                sqlite3_exec(client->dstore.handle, "DETACH mail;", NULL, NULL, NULL);

                DStoreTranAbort(client);
                break;
            }

            if (ccode == SQLITE_OK) {
                ccode = DStoreTranEnd(client);
                if (ccode == SQLITE_DONE) {
                    ccode = SQLITE_OK;
                }
            } else {
                XplConsolePrintf("NMAP [%d]: Upgrade mail folder \"%s\" for \"%s\"; update error \"%s\".\r\n", __LINE__, client->mail.name, client->user, sqlite3_errmsg(client->dstore.handle));

                DStoreTranAbort(client);
            }

            sqlite3_exec(client->dstore.handle, "DETACH mail;", NULL, NULL, NULL);
            break;
        }

        DStoreClose(client);

        ReadNLockRelease(client->mail.lock);
        client->mail.lock = NULL;
    } else {
        ccode = SQLITE_LOCKED;
        XplConsolePrintf("NMAP [%d]: Upgrade mail folder \"%s\" for \"%s\"; locked.\r\n", __LINE__, client->mail.name, client->user);
    }

    client->mail.lock = lock;

    return(ccode);
}

static int 
UpgradeNmapCalStore(NMAPClient *client)
{
    return(SQLITE_OK);
}

static int 
UpgradeNmapCavStore(NMAPClient *client)
{
    int ccode;
    unsigned char line[XPL_MAX_PATH + 24];
    unsigned char guid[XPL_MAX_PATH + 1];
    NLockStruct *lock = client->mail.lock;

    client->cal.lock = NULL;
    GuidAlloc(guid);
    guid[NMAP_GUID_LENGTH] = '\0';

    if (ReadNLockAquire(&client->cal.lock, &client->userHash, client->user, client->cal.name, 3000) == NLOCK_AQUIRED) {
        sprintf(client->path, "%s/%s/nmap.db", client->store, client->user);
        ccode = sqlite3_open(client->path, &client->dstore.handle);

        if (ccode == SQLITE_OK) {
            sprintf(line, "INSERT INTO map (guid, type, start, toDoc, length, resource, indexed, modified) VALUES ('%s', %d, 0, 0, 0, '%s', 0, julianday(DATETIME('now')));", guid, NMAP_DOCTYPE_CAL_FOLDER, client->cal.name);
            ccode = sqlite3_exec(client->dstore.handle, line, NULL, NULL, NULL);
        }

        DStoreClose(client);

        ReadNLockRelease(client->cal.lock);
        client->cal.lock = NULL;
    } else {
        ccode = SQLITE_LOCKED;
        XplConsolePrintf("NMAP [%d]: Upgrade CalDAV folder \"%s\" for \"%s\"; locked.\r\n", __LINE__, client->cal.name, client->user);
    }

    client->cal.lock = lock;

    return(ccode);
}

static int 
UpgradeNmapStore(NMAPClient *client, unsigned char *path, int remaining, int offset)
{
    int ccode = NMAP_OK;
    int length;
    int count;
    XplDir *dirEntryP;
    XplDir *dirP;

    /*  path                                        offset
        /Program Files/Hula/users/admin             0
        /Program Files/Hula/users/admin/Work        -6
        /Program Files/Hula/users/admin/Work/2004   -10
    */

    XplSetThreadGroupID(NMAP.id.load);

    length = strlen(path);

    dirEntryP = XplOpenDir(path);
    if (dirEntryP) {
        while ((dirP = XplReadDir(dirEntryP)) != NULL) {
            if (dirP->d_attr & XPL_A_SUBDIR) {
                if (dirP->d_name[0] != '.') {
                    count = strlen(dirP->d_name);

                    path[length++] = '/';
                    memcpy(path + length, dirP->d_name, count);
                    path[length + count] = '\0';

                    ccode = UpgradeNmapStore(client, path, remaining - count - 1, offset - count - 1);

                    path[--length] = '\0';

                    if (ccode != SQLITE_OK) {
                        break;
                    }
                }

                continue;
            }

            count = strlen(dirP->d_name) - 4;
            if (XplStrCaseCmp(dirP->d_name + count, ".box") == 0) {
                path[length++] = '/';
                memcpy(path + length, dirP->d_name, count);
                path[length + count] = '\0';

                strcpy(client->mail.name, path + length + offset);
                ccode = UpgradeNmapMailStore(client);

                path[--length] = '\0';
            } else if (XplStrCaseCmp(dirP->d_name + count, ".cal") == 0) {
                path[length++] = '/';
                memcpy(path + length, dirP->d_name, count);
                path[length + count] = '\0';

                strcpy(client->cal.name, path + length + offset);
                ccode = UpgradeNmapCalStore(client);

                path[--length] = '\0';
            } else if (XplStrCaseCmp(dirP->d_name + count, NMAP_CAL_EXT) == 0) {
                path[length++] = '/';
                memcpy(path + length, dirP->d_name, count);
                path[length + count] = '\0';

                strcpy(client->cal.name, path + length + offset);
                ccode = UpgradeNmapCavStore(client);

                path[--length] = '\0';
            } else {
                continue;
            }

            if (ccode == SQLITE_OK) {
                continue;
            }

            break;
        }
        XplCloseDir(dirEntryP);
    } else {
        ccode = SQLITE_IOERR;
    }

    XplSetThreadGroupID(NMAP.id.group);

    return(ccode);
}

static int 
DStoreCreate(NMAPClient *client)
{
    int ccode;
    int count = 30;

    sprintf(client->path, "%s/%s/nmap.db", client->store, client->user);
    ccode = sqlite3_open(client->path, &client->dstore.handle);
    if (ccode == SQLITE_OK) {
        do {
            ccode = sqlite3_exec(client->dstore.handle, NMAP_SQLTRAN_BEGIN, NULL, NULL, NULL);
            if (ccode == SQLITE_OK) {
                break;
            }

            if (ccode == SQLITE_BUSY) {
                XplDelay(333);

                count--;
                continue;
            }
        } while ((ccode == SQLITE_BUSY) && count);
    } else {
        XplConsolePrintf("NMAP [%d]: Open dstore database for \"%s\" failed.\r\n", __LINE__, client->user);
        return(ccode);
    }

    if (ccode == SQLITE_OK) {
        ccode = sqlite3_exec(client->dstore.handle, "CREATE TABLE control (" \
                                                    "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
                                                    "version INTEGER NOT NULL, " \
                                                    "modified DATETIME NOT NULL" \
                                                    ");", NULL, NULL, NULL);
        if (ccode == SQLITE_OK) {
            sprintf(client->line, "INSERT INTO control (version, modified) VALUES (%d, julianday(DATETIME('now')));", NMAP_DSTORE_VERSION);
            ccode = sqlite3_exec(client->dstore.handle, client->line, NULL, NULL, NULL);
        } else {
            XplConsolePrintf("NMAP [%d]: Create dstore control table for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            ccode = sqlite3_exec(client->dstore.handle, "CREATE TABLE map (" \
                                                        "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
                                                        "guid VARCHAR(46) NOT NULL UNIQUE, " \
                                                        "flags INTEGER DEFAULT 0, " \
                                                        "type INTEGER DEFAULT 0, " \
                                                        "start INTEGER NOT NULL, " \
                                                        "toDoc INTEGER NOT NULL, " \
                                                        "length INTEGER NOT NULL, " \
                                                        "resource TEXT DEFAULT NULL, " \
                                                        "indexed INTEGER DEFAULT 0, " \
                                                        "modified DATETIME NOT NULL" \
                                                        ");", NULL, NULL, NULL);
        } else {
            XplConsolePrintf("NMAP [%d]: Create dstore resource index for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            ccode = sqlite3_exec(client->dstore.handle, "CREATE INDEX map_index ON map (guid, indexed, modified);", NULL, NULL, NULL);
        } else {
            XplConsolePrintf("NMAP [%d]: Create dstore map table for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            ccode = sqlite3_exec(client->dstore.handle, "CREATE TABLE prop (" \
                                                        "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
                                                        "guid VARCHAR(46) NOT NULL, " \
                                                        "name TEXT NOT NULL, " \
                                                        "value TEXT NOT NULL" \
                                                        ");", NULL, NULL, NULL);
        } else {
            XplConsolePrintf("NMAP [%d]: Create dstore map index for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            CreatePurgedTable(client);
        } else {
            XplConsolePrintf("NMAP [%d]: Create purged table for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            ccode = sqlite3_exec(client->dstore.handle, "CREATE INDEX prop_index ON prop (guid, name);", NULL, NULL, NULL);
            if (ccode != SQLITE_OK) {
                XplConsolePrintf("NMAP [%d]: Create dstore prop index for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
            }
        } else {
            XplConsolePrintf("NMAP [%d]: Create dstore prop table for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_OK) {
            ccode = sqlite3_exec(client->dstore.handle, NMAP_SQLTRAN_END, NULL, NULL, NULL);
        } else {
            sqlite3_exec(client->dstore.handle, NMAP_SQLTRAN_ROLLBACK, NULL, NULL, NULL);
            sqlite3_exec(client->dstore.handle, NMAP_SQLTRAN_END, NULL, NULL, NULL);
        }
    } else {
        XplConsolePrintf("NMAP [%d]: Begin dstore exclusive transaction for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
    }

    DStoreClose(client);

    return(ccode);
}

void
CreatePurgedTable(NMAPClient *client)
{
    sqlite3_exec(client->dstore.handle, "CREATE TABLE purged (" \
                 "id INTEGER PRIMARY KEY AUTOINCREMENT, " \
                 "guid VARCHAR(46) NOT NULL" \
                 ");", NULL, NULL, NULL);
}

void 
DStoreClose(NMAPClient *client)
{
    unsigned int i;
    sqlite3_stmt **stmt;

    if (client->dstore.handle) {
        stmt = &(client->dstore.sql.begin);
        for (i = 0; i < (sizeof(NmapClientDStoreStatements) / sizeof(sqlite3_stmt *)); i++, stmt++) {
            if (*stmt) {
                sqlite3_finalize(*stmt);
                *stmt = NULL;
            }
        }

        sqlite3_close(client->dstore.handle);
        client->dstore.handle = NULL;
    }

    return;
}

int
DStoreOpen(NMAPClient *client)
{
    int ccode;
    int remaining = XPL_MAX_PATH;
    int version;
    unsigned char *path = NULL;
    sqlite3_stmt *stmt;

    if (client->dstore.handle) {
        return(SQLITE_DONE);
    }

    path = MemMalloc(remaining + 1);
    if (path) {
        if (ReadNLockAquire(&client->mail.lock, &client->userHash, client->user, "", 3000) == NLOCK_AQUIRED) {
            if (PurgeNLockAquire(client->mail.lock, 750) == NLOCK_AQUIRED) {
                ccode = SQLITE_OK;
            } else {
                ccode = SQLITE_BUSY;
                ReadNLockRelease(client->mail.lock);
                client->mail.lock = NULL;
            }
        } else {
            ccode = SQLITE_BUSY;
        }
    } else {
        return(SQLITE_NOMEM);
    }

    if (ccode == SQLITE_OK) {
        sprintf(client->path, "%s/%s/nmap.db", client->store, client->user);
        if (access(client->path, 0)) {
            ccode = DStoreCreate(client);
            if (ccode == SQLITE_OK) {

                remaining -= sprintf(path, "%s/%s", client->store, client->user);
                ccode = UpgradeNmapStore(client, path, remaining, 0);
            } else {
                XplConsolePrintf("NMAP [%d]: Upgrade mail folder \"%s\" for \"%s\"; prepare error \"%s\".\r\n", __LINE__, client->mail.name, client->user, sqlite3_errmsg(client->dstore.handle));
            }
        }
    } else {
        MemFree(path);

        return(ccode);
    }

    if (ccode == SQLITE_OK) {
        ccode = sqlite3_open(client->path, &client->dstore.handle);
        if (ccode == SQLITE_OK) {
            ccode = DStoreTranBegin(client, TRUE);
        } else {
            XplConsolePrintf("NMAP [%d]: Open dstore database for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
        }

        if (ccode == SQLITE_DONE) {
            ccode = sqlite3_prepare(client->dstore.handle, "SELECT version FROM control;", 28, &stmt, NULL);
            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    version = sqlite3_column_int(stmt, 0);
                    if (version == NMAP_DSTORE_VERSION) {
                        ccode = SQLITE_DONE;
                    } else {
                        XplConsolePrintf("NMAP [%d]: Unknown dstore control version of %d for \"%s\"\r\n", __LINE__, version, client->user);

                        ccode = SQLITE_MISMATCH;
                    }
                } else if (ccode == SQLITE_DONE) {
                    XplConsolePrintf("NMAP [%d]: Select dstore version for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
                }

                sqlite3_finalize(stmt);
            } else {
                XplConsolePrintf("NMAP [%d]: Prepare select dstore version for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(client->dstore.handle));
            }

            CreatePurgedTable(client);

            if (ccode == SQLITE_DONE) {
                ccode = DStoreTranEnd(client);
            } else {
                DStoreTranAbort(client);
            }
        }

        if (ccode != SQLITE_DONE) {
            DStoreClose(client);
        }
    }

    PurgeNLockRelease(client->mail.lock);
    ReadNLockRelease(client->mail.lock);
    client->mail.lock = NULL;

    MemFree(path);

    return(ccode);
}

static int 
SendVerboseMessageStateFlags(NMAPClient *client, MessageStateFlags state)
{
    int ccode;
    unsigned char *ptr;
    unsigned char buffer[112]; /* Combinded length of all possible flags + 1 space per flag. */
    MessageStateFlags mask;

    mask = MSG_STATE_READ \
            | MSG_STATE_DELETED \
            | MSG_STATE_NEWUIDL \
            | MSG_STATE_ESCAPE \
            | MSG_STATE_PRIOHIGH \
            | MSG_STATE_PRIOLOW \
            | MSG_STATE_PURGED \
            | MSG_STATE_RECENT \
            | MSG_STATE_ANSWERED \
            | MSG_STATE_DRAFT \
            | MSG_STATE_MARK_PURGE \
            | MSG_STATE_COMPLETED \
            | MSG_STATE_PRIVATE;

    if (state & mask) {
        state &= mask;

        ptr = buffer;
    } else {
        return(ConnWrite(client->conn, "0\r\n", 3));
    }

    if (state & MSG_STATE_READ) {
        state &= ~MSG_STATE_READ;
        ptr += sprintf(ptr, "Read%s", state? " ": "");
    }

    if (state & MSG_STATE_DELETED) {
        state &= ~MSG_STATE_DELETED;
        ptr += sprintf(ptr, "Deleted%s", state? " ": "");
    }

    if (state & MSG_STATE_NEWUIDL) {
        state &= ~MSG_STATE_NEWUIDL;
        ptr += sprintf(ptr, "NewUIDL%s", state? " ": "");
    }

    if (state & MSG_STATE_ESCAPE) {
        state &= ~MSG_STATE_ESCAPE;
        ptr += sprintf(ptr, "Escape%s", state? " ": "");
    }

    if (state & MSG_STATE_PRIOHIGH) {
        state &= ~MSG_STATE_PRIOHIGH;
        ptr += sprintf(ptr, "PriorityHigh%s", state? " ": "");
    }

    if (state & MSG_STATE_PRIOLOW) {
        state &= ~MSG_STATE_PRIOLOW;
        ptr += sprintf(ptr, "PriorityLow%s", state? " ": "");
    }

    if (state & MSG_STATE_PURGED) {
        state &= ~MSG_STATE_PURGED;
        ptr += sprintf(ptr, "Purged%s", state? " ": "");
    }

    if (state & MSG_STATE_RECENT) {
        state &= ~MSG_STATE_RECENT;
        ptr += sprintf(ptr, "Recent%s", state? " ": "");
    }

    if (state & MSG_STATE_ANSWERED) {
        state &= ~MSG_STATE_ANSWERED;
        ptr += sprintf(ptr, "Answered%s", state? " ": "");
    }

    if (state & MSG_STATE_DRAFT) {
        state &= ~MSG_STATE_DRAFT;
        ptr += sprintf(ptr, "Draft%s", state? " ": "");
    }

    if (state & MSG_STATE_MARK_PURGE) {
        state &= ~MSG_STATE_MARK_PURGE;
        ptr += sprintf(ptr, "MarkedPurge%s", state? " ": "");
    }

    if (state & MSG_STATE_COMPLETED) {
        state &= ~MSG_STATE_COMPLETED;
        ptr += sprintf(ptr, "Completed%s", state? " ": "");
    }

    if (state & MSG_STATE_PRIVATE) {
        state &= ~MSG_STATE_PRIVATE;
        ptr += sprintf(ptr, "Private%s", state? " ": "");
    }

    if (state & MSG_STATE_NOFREEBUSY) {
        state &= ~MSG_STATE_NOFREEBUSY;
        ptr += sprintf(ptr, "NoFreeBusy%s", state? " ": "");
    }

    if (state & MSG_STATE_USE_SCMS) {
        state &= ~MSG_STATE_USE_SCMS;
        ptr += sprintf(ptr, "UseSCMS%s", state? " ": "");
    }

    if (state & MSG_STATE_MODIFIED) {
        state &= ~MSG_STATE_MODIFIED;
        ptr += sprintf(ptr, "Modified%s", state? " ": "");
    }

    if ((ccode = ConnWriteF(client->conn, "%d\r\n", ptr - buffer)) != -1) {
        ccode = ConnWrite(client->conn, buffer, ptr - buffer);
    }

    return(ccode);
}

static unsigned int 
WriteDocumentHeaders(NMAPClient *client, FILE *fh, unsigned char *guid, int size)
{
    int count;
    size_t length;
    time_t tod;

    time(&tod);

    length = sprintf(client->line, "From %s ", client->user);
    count = length + 3;

    length += strftime(client->line + length, sizeof(client->line) - length, "%a %b %d %H:%M %Y", gmtime(&tod));

    if (client->line[count] == ' ') {
        client->line[length++] = CR;
        client->line[length++] = LF;
    } else {
        length = strlen(client->line + count + 1);
        memmove(client->line + count, client->line + count + 1, length);
        length += count;

        client->line[length++] = CR;
        client->line[length++] = LF;
    }

    if (fwrite(client->line, sizeof(unsigned char), length, fh) == length) {
        count = fprintf(fh, "nmap.guid=%s\r\nnmap.length=%d\r\n\r\n", guid, size);
        if (count != -1) {
            return(length + count);
        }
    }

    return(-1);
}

int 
NmapCommandDcopy(void *param)
{
    int ccode;
    NMAPClient *client = (NMAPClient *)param;

    ccode = ConnWrite(client->conn, MSG1000NOTIMP, sizeof(MSG1000NOTIMP) - 1);

    return(ccode);
}

int 
NmapCommandDcrea(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *cur;
    unsigned char scratch[(2 * XPL_MAX_PATH) + 1];
    char guid[NMAP_GUID_LENGTH+1];
    struct stat sb;
    FILE *fh = NULL;
    sqlite3_stmt *stmt;
    sqlite3 *handle;
    XplDir *dirP;
    XplDir *dirEntryP;
    NLockStruct *lock = NULL;
    NmapDStoreInfo info;
    NMAPClient *client = (NMAPClient *)param;

    /* DCREA <Resource> <Type> [<NMAP GUID>] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            memset(&info, 0, sizeof(info));

            ptr = client->buffer + 5;

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if ((*ptr++ == ' ') 
            && *ptr 
            && (*ptr != ' ') 
            && (*ptr != '.') 
            && ((cur = strchr(ptr, ' ')) != NULL)) {
        *cur++ = '\0';

        ptr = strchr(cur, ' ');
        if (ptr) {
            *ptr++ = ' ';
            if ((strchr(ptr, ' ') == NULL) && (strlen(ptr) == NMAP_GUID_LENGTH)) {
                memcpy(info.guid, ptr, NMAP_GUID_LENGTH + 1);
            } else {
                return(ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1));
            }
        }

        strcpy(info.resource, client->buffer + 6);
        info.type = atoi(cur);
        if (NMAP_IS_FOLDER(info.type)
            && (strstr(info.resource, "../") == NULL) 
            && (strstr(info.resource, "..\\") == NULL)) 
        {
            cur = strchr(info.resource, '.');
            if (cur) {
                if (info.type) {
                    return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
                }
            } else if (!info.type) {
                if (!cur || (cur[1] == '\0')) {
                    return(ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal mailbox name."));
                }
            }
        } else {
            return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (((ccode = NmapSqlPrepare(handle, NMAP_DINFO_INSERT, &(client->dstore.sql.ins.dstor))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK)) {
        stmt = client->dstore.sql.sel.dstor;
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (info.guid[0] == '\0') {
        GuidAlloc(info.guid);
        info.guid[NMAP_GUID_LENGTH] = '\0';
    } else if ((ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) {
        ccode = sqlite3_step(stmt);
        if (ccode == SQLITE_DONE) {
            ccode = 0;
        } else {
            stmt = NULL;

            if (ccode == SQLITE_ROW) {
                ccode = ConnWriteF(client->conn, "4226 Guid %s already exists\r\n", info.guid);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        }

        sqlite3_reset(client->dstore.sql.sel.dstor);

        if (!stmt) {
            return(ccode);
        }
    } else {
        sqlite3_reset(stmt);
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    cur = client->path + sprintf(client->path, "%s/%s/", client->store, client->user);
    if ((ReadNLockAquire(&lock, &client->userHash, client->user, info.resource, 3000) == NLOCK_AQUIRED) 
            && (WriteNLockAquire(lock, 2000) == NLOCK_AQUIRED)) {
        strcpy(cur, info.resource);
        ptr = cur;

        ccode = FindResourceGuid(client, cur, info.type, guid);
        if (!ccode) {
            ccode = NmapCheckAuthorization(client, guid, NMAP_PRIV_BIND, 0);
        } else if (-4224 == ccode) {
            /* user is the only one who can create top-level resources */
            if (NMAP_PRINCIPAL_USER != client->principalType || 
                (client->principal[0] && strcmp(client->principal, client->user)))
            {
                ccode = ConnWrite(client->conn, MSG4240NOPERMISSION, 
                                  sizeof(MSG4240NOPERMISSION) - 1);
            } else {
                ccode = 0;
            }
        }
        if (ccode) {
            WriteNLockRelease(lock);
            ReadNLockRelease(lock);
            return ccode;
        }

        while ((ptr = strchr(ptr, '/'))) {
            *ptr = '\0';
            if (stat(client->path, &sb) != 0) {
                XplMakeDir(client->path);
                SetLongName(client->path, scratch);
            }
            *ptr++ = '/';
        }
    } else {
        if (lock) {
            ReadNLockRelease(lock);
            lock = NULL;
        }

        return(ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1));
    }


    do {
        switch (info.type) {
            case NMAP_DOCTYPE_UNKNOWN_FOLDER: {
                fh = fopen(client->path, "r");
                break;
            }

            case NMAP_DOCTYPE_MAIL_FOLDER: {
                strcat(cur, NMAP_MAIL_EXT);
                fh = fopen(client->path, "r");
                break;
            }

            case NMAP_DOCTYPE_CAL_FOLDER: {
                strcat(cur, NMAP_CAL_EXT);
                fh = fopen(client->path, "r");
                break;
            }

            case NMAP_DOCTYPE_AB_FOLDER: {
                strcat(cur, NMAP_ADBK_EXT);
                fh = fopen(client->path, "r");
            }

            case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
            case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
            case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
            case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
            case NMAP_DOCTYPE_UNKNOWN: 
            case NMAP_DOCTYPE_MAIL: 
            case NMAP_DOCTYPE_CAL: 
            case NMAP_DOCTYPE_AB: 
            case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: 
            case NMAP_DOCTYPE_MAIL_IN_SHARE: 
            case NMAP_DOCTYPE_CAL_IN_SHARE: 
            case NMAP_DOCTYPE_AB_IN_SHARE: 
            case NMAP_DOCTYPE_PURGED:
            default: {
                fh = NULL;
                break;
            }
        }

        if (!fh && (IsAvailableShare(client->user, client->path, NMAP_SHARE_MAILBOX, NULL, NULL) == FALSE)) {
            cur--;
            *cur = '\0';
            dirEntryP = XplOpenDir(client->path);
        } else {
            fclose(fh);
            fh = NULL;

            ccode = ConnWrite(client->conn, MSG4226MBOXEXISTS, sizeof(MSG4226MBOXEXISTS) - 1);
            break;
        }

        if (dirEntryP != NULL) {
            while ((dirP = XplReadDir(dirEntryP)) != NULL) {
                if (dirP->d_attr & XPL_A_SUBDIR) {
                    continue;
                }

                ptr = strchr(dirP->d_name, '.');
                if (ptr) {
                    switch (info.type) {
                        case NMAP_DOCTYPE_UNKNOWN_FOLDER: {
                            break;
                        }

                        case NMAP_DOCTYPE_MAIL_FOLDER: {
                            if ((XplStrCaseCmp(ptr, NMAP_MAIL_INDEX_EXT) == 0) 
                                    || (XplStrCaseCmp(ptr, ".idx") == 0) 
                                    || (XplStrCaseCmp(ptr, ".sdx") == 0)) {
                                sprintf(cur, "/%s", dirP->d_name);
                                unlink(client->path);

                                *cur = '\0';
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_CAL_FOLDER: {
                            if ((XplStrCaseCmp(ptr, NMAP_CAL_INDEX_EXT) == 0) 
                                    || (XplStrCaseCmp(ptr, ".idc") == 0) 
                                    || (XplStrCaseCmp(ptr, ".sdc") == 0)) {
                                sprintf(cur, "/%s", dirP->d_name);
                                unlink(client->path);

                                *cur = '\0';
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_AB_FOLDER: {
                            if (XplStrCaseCmp(ptr, NMAP_ADBK_INDEX_EXT) == 0) {
                                sprintf(cur, "/%s", dirP->d_name);
                                unlink(client->path);

                                *cur = '\0';
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_UNKNOWN: 
                        case NMAP_DOCTYPE_MAIL: 
                        case NMAP_DOCTYPE_CAL: 
                        case NMAP_DOCTYPE_AB: 
                        case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: 
                        case NMAP_DOCTYPE_MAIL_IN_SHARE: 
                        case NMAP_DOCTYPE_CAL_IN_SHARE: 
                        case NMAP_DOCTYPE_AB_IN_SHARE: 
                        case NMAP_DOCTYPE_PURGED: 
                        default: {
                            break;
                        }
                    }
                }
            }
            XplCloseDir(dirEntryP);
        }

        *cur++ = '/';
        switch (info.type) {
            case NMAP_DOCTYPE_UNKNOWN_FOLDER: {
                sprintf(client->path, "%s/%s/%s", client->store, client->user, info.resource);
                fh = fopen(client->path, "wb");
                if (fh) {
                    fclose(fh);

                    SetLongName(client->path, scratch);
                } else {
                    ccode = ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal name.");
                }

                break;
            }

            case NMAP_DOCTYPE_MAIL_FOLDER: {
                sprintf(client->path, "%s/%s/%s"NMAP_MAIL_EXT, client->store, client->user, info.resource);
                fh = fopen(client->path, "wb");
                if (fh) {
                    fclose(fh);

                    SetLongName(client->path, scratch);

                    strcpy(client->mail.name, info.resource);
                    if (MailInfoOpen(client) == SQLITE_OK) {
                        MailInfoClose(client);
                        break;
                    }
                } else {
                    ccode = ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal mailbox name.");
                    break;
                }

                fh = NULL;

                sprintf(client->path, "%s/%s/%s"NMAP_MAIL_EXT, client->store, client->user, info.resource);
                unlink(client->path);

                sprintf(client->path, "%s/%s/%s"NMAP_MAIL_INDEX_EXT, client->store, client->user, info.resource);
                unlink(client->path);

                sprintf(client->path, "%s/%s/%s.idx", client->store, client->user, info.resource);
                unlink(client->path);

                ccode = ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal mailbox name.");
                break;
            }

            case NMAP_DOCTYPE_CAL_FOLDER: {
                sprintf(client->path, "%s/%s/%s"NMAP_CAL_EXT, client->store, client->user, info.resource);
                fh = fopen(client->path, "wb");
                if (fh) {
                    fclose(fh);

                    SetLongName(client->path, scratch);
                } else {
                    ccode = ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal calendar name.");
                }

                break;
            }

            case NMAP_DOCTYPE_AB_FOLDER: {
                sprintf(client->path, "%s/%s/%s"NMAP_ADBK_EXT, client->store, client->user, info.resource);
                fh = fopen(client->path, "wb");
                if (fh) {
                    fclose(fh);

                    SetLongName(client->path, scratch);
                } else {
                    ccode = ConnWriteF(client->conn, MSG3014REQNOTALLOWED, " Illegal address book name.");
                }

                break;
            }

            case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER:
            case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
            case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
            case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
            case NMAP_DOCTYPE_UNKNOWN:
            case NMAP_DOCTYPE_MAIL: 
            case NMAP_DOCTYPE_CAL: 
            case NMAP_DOCTYPE_AB: 
            case NMAP_DOCTYPE_UNKNOWN_IN_SHARE:
            case NMAP_DOCTYPE_MAIL_IN_SHARE: 
            case NMAP_DOCTYPE_CAL_IN_SHARE: 
            case NMAP_DOCTYPE_AB_IN_SHARE: 
            case NMAP_DOCTYPE_PURGED:
            default: {
                fh = NULL;
                break;
            }
        }

        if (fh) {
            fh = NULL;

            ccode = DStoreTranBegin(client, TRUE);
            if (ccode == SQLITE_DONE) {
                ;
            } else {
                ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
                break;
            }
        } else {
            break;
        }

        stmt = client->dstore.sql.ins.dstor;
        if (((ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_int(stmt, 2, info.type)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_int(stmt, 3, 0)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_int(stmt, 4, 0)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_int(stmt, 5, 0)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_text(stmt, 6, info.resource, -1, SQLITE_STATIC)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_int(stmt, 7, 0)) == SQLITE_OK)) {
            ccode = sqlite3_step(stmt);
        } else {
            stmt = NULL;
        }

        sqlite3_reset(client->dstore.sql.ins.dstor);

        if (stmt && (ccode == SQLITE_DONE)) {
            stmt = client->dstore.sql.sel.dstor;
            if ((ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    strcpy(info.version.store, sqlite3_column_text(stmt, 0));
                } else {
                    stmt = NULL;
                }
            } else {
                stmt = NULL;
            }

            sqlite3_reset(client->dstore.sql.sel.dstor);
        } else {
            stmt = NULL;
        }

        if (stmt) {
            ccode = DStoreTranEnd(client);
            if (ccode == SQLITE_DONE) {
                ccode = ConnWriteF(client->conn, "1000 %s %s Mailbox created\r\n", info.guid, info.version.store);
                break;
            }
        }

        DStoreTranAbort(client);

        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
        break;
    } while (TRUE);

    WriteNLockRelease(lock);
    ReadNLockRelease(lock);

    return(ccode);
}

int 
NmapCommandDguid(void *param)
{
    int i;
    int ccode = 0;
    int count;
    unsigned char *ptr;
    NMAPClient *client = (NMAPClient *)param;

    /* DGUID [<count>] */
    if (client->states & NMAP_CLIENT_USER) {
        ptr = client->buffer + 5;
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        count = 1;
    } else if (*ptr == ' ') {
        ptr++;

        if (isdigit(*ptr)) {
            count = atoi(ptr);
        } else {
            return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
        }

        if (count > 65535) {
            return(ConnWrite(client->conn, "3010 Count exceed allowed limit.\r\n", 34));
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    memset(client->line, '0', NMAP_GUID_LENGTH + 7);
    ptr = client->line;

    ptr[0] = '2';
    ptr[3] = '1';
    ptr[4] = ' ';
    ptr[NMAP_GUID_LENGTH + 5] = CR;
    ptr[NMAP_GUID_LENGTH + 6] = LF;

    ptr += 5;
    for (i = 0; i < count; i++) {
        GuidAlloc(ptr);

        if ((ccode = ConnWrite(client->conn, client->line, NMAP_GUID_LENGTH + 7)) != -1) {
            continue;
        }

        break;
    }

    if (ccode != -1) {
        ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
    }

    return(ccode);
}

int 
NmapCommandDinfo(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *guid;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle;
    sqlite3_stmt *stmt;

    /* DINFO [*(<NMAP GUID>)] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 5;

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        ccode = NmapSqlPrepare(handle, NMAP_DSTORE_SELECT, &(client->dstore.sql.sel.dinfo));
        if (ccode == SQLITE_OK) {
            ccode = DStoreTranBegin(client, FALSE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode == SQLITE_DONE) {
            stmt = client->dstore.sql.sel.dinfo;

            while ((ccode = sqlite3_step(stmt)) == SQLITE_ROW) {
                char iguid[NMAP_GUID_LENGTH+1];

                strncpy(iguid, sqlite3_column_text(stmt, 0), NMAP_GUID_LENGTH);
                iguid[NMAP_GUID_LENGTH] = 0;
                if (NmapCheckAuthorization(client, iguid, NMAP_PRIV_READ, 1)) {
                    continue;
                }

                if ((ccode = ConnWriteF(client->conn, "2001 %s %s %d %d %s\r\n", 
                        iguid,
                        sqlite3_column_text(stmt, 1), 
                        sqlite3_column_int(stmt, 2), 
                        sqlite3_column_int(stmt, 3), 
                        sqlite3_column_text(stmt, 4))) != -1) {
                    continue;
                }

                break;
            }

            if (ccode == SQLITE_DONE) {
                ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);

            DStoreTranEnd(client);
        } else {
            ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
        }

        return(ccode);
    }

    if (*ptr == ' ') {
        ccode = NmapSqlPrepare(handle, NMAP_DINFO_BY_GUID, &(client->dstore.sql.sel.dinfoGuid));
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (ccode == SQLITE_OK) {
        ccode = DStoreTranBegin(client, FALSE);
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (ccode == SQLITE_DONE) {
        stmt = client->dstore.sql.sel.dinfoGuid;

        ccode = ConnWrite(client->conn, "2000 Processing\r\n", 17);
        while (ccode != -1) {
            if (ptr) {
                guid = ++ptr;
                ptr = strchr(ptr, ' ');
            } else {
                break;
            }

            if (ptr) {
                *ptr = '\0';

                if ((ptr - guid) == NMAP_GUID_LENGTH) {
                    ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                    ccode = SQLITE_DONE;
                    continue;
                } else {
                    break;
                }
            } else if (strlen(guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                break;
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    if (!NmapCheckAuthorization(client, guid, NMAP_PRIV_READ, 0)) {
                        ccode = ConnWriteF(client->conn, "2001 %s %s %d %d %s\r\n", 
                                           guid, 
                                           sqlite3_column_text(stmt, 0), 
                                           sqlite3_column_int(stmt, 1), 
                                           sqlite3_column_int(stmt, 5), 
                                           sqlite3_column_text(stmt, 6));
                    }
                } else if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                } else {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);
        }

        DStoreTranEnd(client);

        if (ccode != -1) {
            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
        }
    } else {
        ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
    }

    return(ccode);
}

int 
NmapCommandDinset(void *param)
{
    int ccode;
    int value;
    unsigned char *ptr;
    unsigned char *guid;
    BOOL commit;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt;

    /* DINSET <Value> *(<NMAP GUID>) */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = strchr(client->buffer + 6, ' ');
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (ptr++ && isdigit(*ptr) && ((value = atoi(ptr)) > -1) && ((ptr = strchr(ptr, ' ')) != NULL)) {
        ccode = NmapSqlPrepare(handle, NMAP_DINSET_UPDATE, &(client->dstore.sql.upd.dinset));
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (ccode == SQLITE_OK) {
        ccode = DStoreTranBegin(client, TRUE);
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (ccode == SQLITE_DONE) {
        stmt = client->dstore.sql.upd.dinset;
        commit = TRUE;

        ccode = ConnWrite(client->conn, "2000 Processing\r\n", 17);
        while (ccode != -1) {
            if (ptr) {
                guid = ++ptr;
                ptr = strchr(ptr, ' ');
            } else {
                break;
            }

            if (ptr) {
                *ptr = '\0';

                if ((ptr - guid) == NMAP_GUID_LENGTH) {
                    ccode = sqlite3_bind_int(stmt, 1, value);
                } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                    ccode = SQLITE_DONE;
                    continue;
                } else {
                    commit = FALSE;
                    break;
                }
            } else if (strlen(guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_int(stmt, 1, value);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                commit = FALSE;
                break;
            }

            if (NmapCheckAuthorization(client, guid, NMAP_PRIV_WRITE, 0)) {
                sqlite3_reset(stmt);
                continue;
            }
                
            if ((ccode == SQLITE_OK) 
                    && ((ccode = sqlite3_bind_text(stmt, 2, guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK)) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_DONE) {
                    ccode = ConnWriteF(client->conn, "2001 %s is %d\r\n", guid, value);

                    sqlite3_reset(stmt);
                    continue;
                }

                XplConsolePrintf("NMAP [%d]: Update dinset for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(handle));
            }

            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            sqlite3_reset(stmt);
        }
    } else {
        commit = FALSE;
    }

    if (ccode != -1) {
        if (commit) {
            ccode = DStoreTranEnd(client);
            if (ccode == SQLITE_DONE) {
                return(ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1));
            }
        } else {
            DStoreTranAbort(client);
        }

        ccode = ConnWrite(client->conn, MSG5004INTERNALERR, sizeof(MSG5004INTERNALERR) - 1);
    }

    return(ccode);
}

static int 
DStoreListFolder(NMAPClient *client, NmapDStoreInfo *info)
{
    int ccode;
    int count;
    NmapDocumentTypes type;
    sqlite3_stmt *stmt;
    sqlite3_stmt *stmtCount;

    if ((ccode = NmapSqlPrepare(client->dstore.handle, NMAP_DLIST_FOLDER_COUNT, &(client->dstore.sql.sel.dlistCount))) == SQLITE_OK) {
        ccode = NmapSqlPrepare(client->dstore.handle, NMAP_DLIST_FOLDER, &(client->dstore.sql.sel.dlistFolder));
    }

    if (ccode == SQLITE_OK) {
        stmt = client->dstore.sql.sel.dlistFolder;
        stmtCount = client->dstore.sql.sel.dlistCount;

        type = info->type & ~NMAP_DOCTYPE_FOLDER;

        if (((ccode = sqlite3_bind_text(stmtCount, 1, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)
                && ((ccode = sqlite3_bind_int(stmtCount, 2, type)) == SQLITE_OK) 
                && ((ccode = sqlite3_bind_text(stmt, 1, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)
                && ((ccode = sqlite3_bind_int(stmt, 2, type)) == SQLITE_OK)) {
            if ((ccode = sqlite3_step(stmtCount)) == SQLITE_ROW) {
                count = sqlite3_column_int(stmtCount, 0);
                ccode = ConnWriteF(client->conn, "2003 %d rows to follow.\r\n", count);

                while ((ccode != -1) 
                        && ((ccode = sqlite3_step(stmt)) == SQLITE_ROW)) {
                    ccode = ConnWriteF(client->conn, "2001 %s %d %s\r\n", 
                            sqlite3_column_text(stmt, 0), 
                            sqlite3_column_int(stmt, 1), 
                            sqlite3_column_text(stmt, 2));
                }

                if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
                } else if (ccode != -1) {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
            } else {
                ccode = ConnWrite(client->conn, "2003 0 rows to follow.\r\n", 24);
            }
        } else {
            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
        }

        sqlite3_reset(stmtCount);
        sqlite3_reset(stmt);
    } else {
        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
    }

    return(ccode);
}

static int 
DStoreListProperties(NMAPClient *client, NmapDStoreInfo *info, BOOL *failure)
{
    int ccode = 0;
    struct stat sb;
    unsigned char *ptr = NULL;
    const char *value;
    sqlite3 *store = NULL;
    sqlite3_stmt *stmt = NULL;

    *failure = FALSE;

    if (((ccode = ConnWriteF(client->conn, "2001 nmap.guid %d\r\n%s", NMAP_GUID_LENGTH, info->guid)) != -1) 
            && ((ccode = ConnWriteF(client->conn, "2001 nmap.type 10\r\n%010d", info->type)) != -1) 
            && ((ccode = ConnWriteF(client->conn, "2001 nmap.resource %d\r\n%s", strlen(info->resource), info->resource)) != -1) 
            && ((ccode = ConnWriteF(client->conn, "2001 nmap.index 10\r\n%010d", info->indexed)) != -1) 
            && ((ccode = ConnWriteF(client->conn, "2001 nmap.modified %d\r\n%s", strlen(info->version.store), info->version.store)) != -1)) {
        ptr = client->path + sprintf(client->path, "%s/%s/%s", client->store, client->user, info->resource);
    }

    switch (info->type) {
        case NMAP_DOCTYPE_UNKNOWN_FOLDER: {
            if (stat(client->path, &sb) == 0) {
                info->length = sb.st_size;
            } else {
                info->length = 0;
            }

            if (ccode != -1) {
                ccode = ConnWriteF(client->conn, "2001 nmap.length 10\r\n%010d", info->length);
            }

            break;
        }

        case NMAP_DOCTYPE_MAIL_FOLDER: {
            memcpy(ptr, NMAP_MAIL_EXT, sizeof(NMAP_MAIL_EXT));

            if (stat(client->path, &sb) == 0) {
                info->length = sb.st_size;
            } else {
                info->length = 0;
            }

            if (ccode != -1) {
                ccode = ConnWriteF(client->conn, "2001 nmap.length 10\r\n%010d", info->length);
            }

            break;
        }

        case NMAP_DOCTYPE_CAL_FOLDER: {
            memcpy(ptr, NMAP_CAL_EXT, sizeof(NMAP_CAL_EXT));

            if (stat(client->path, &sb) == 0) {
                info->length = sb.st_size;
            } else {
                info->length = 0;
            }

            if (ccode != -1) {
                ccode = ConnWriteF(client->conn, "2001 nmap.length 10\r\n%010d", info->length);
            }

            break;
        }

        case NMAP_DOCTYPE_AB_FOLDER: {
            memcpy(ptr, NMAP_ADBK_EXT, sizeof(NMAP_ADBK_EXT));

            if (stat(client->path, &sb) == 0) {
                info->length = sb.st_size;
            } else {
                info->length = 0;
            }

            if (ccode != -1) {
                ccode = ConnWriteF(client->conn, "2001 nmap.length 10\r\n%010d", info->length);
            }

            break;
        }

        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
        case NMAP_DOCTYPE_UNKNOWN: {
            break;
        }

        case NMAP_DOCTYPE_MAIL: {
            memcpy(ptr, NMAP_MAIL_INDEX_EXT, sizeof(NMAP_MAIL_INDEX_EXT));
            if (((ccode = sqlite3_open(client->path, &store)) == SQLITE_OK) 
                    && ((ccode = sqlite3_prepare(store, NMAP_DLIST_MAIL, sizeof(NMAP_DLIST_MAIL) - 1, &stmt, NULL)) == SQLITE_OK) 
                    && ((ccode = sqlite3_bind_text(stmt, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK)) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    if (((ccode = ConnWrite(client->conn, "2001 nmap.state ", 16)) != -1) 
                            && ((ccode = SendVerboseMessageStateFlags(client, sqlite3_column_int(stmt, 0))) != -1) 
                            && ((ccode = ConnWriteF(client->conn, "2001 nmap.uid 10\r\n%010d", sqlite3_column_int(stmt, 1))) != -1) 
                            && ((ccode = ConnWriteF(client->conn, "2001 nmap.sent 10\r\n%010d", sqlite3_column_int(stmt, 2))) != -1)) {
                        ccode = ConnWriteF(client->conn, "2001 nmap.received 10\r\n%010d", sqlite3_column_int(stmt, 3));
                    }
                } else if (ccode == SQLITE_DONE) {
                    *failure = TRUE;
                    ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                } else {
                    *failure = TRUE;
                    ccode = ConnWrite(client->conn, MSG4224CANTREADSQL, sizeof(MSG4224CANTREADSQL) - 1);
                }
            } else {
                    *failure = TRUE;
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            if (stmt) {
                sqlite3_finalize(stmt);
                stmt = NULL;
            }

            if (store) {
                sqlite3_close(store);
                store = NULL;
            }

            memcpy(ptr, NMAP_MAIL_EXT, sizeof(NMAP_MAIL_EXT));
            break;
        }

        case NMAP_DOCTYPE_CAL: {
            memcpy(ptr, NMAP_CAL_EXT, sizeof(NMAP_CAL_EXT));
            break;
        }

        case NMAP_DOCTYPE_AB: {
            memcpy(ptr, NMAP_ADBK_EXT, sizeof(NMAP_ADBK_EXT));
            break;
        }

        case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: 
        case NMAP_DOCTYPE_MAIL_IN_SHARE: 
        case NMAP_DOCTYPE_CAL_IN_SHARE: 
        case NMAP_DOCTYPE_AB_IN_SHARE: {
            break;
        }

        case NMAP_DOCTYPE_PURGED: 
        default: {
            return(0);
        }
    }

    if (((ccode = NmapSqlPrepare(client->dstore.handle, NMAP_DPGET_SELECT, &(client->dstore.sql.sel.dpget))) == SQLITE_OK) 
            && ((ccode = sqlite3_bind_text(client->dstore.sql.sel.dpget, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK)) {
        stmt = client->dstore.sql.sel.dpget;
        while ((ccode = sqlite3_step(stmt)) == SQLITE_ROW) {
            if ((ccode = ConnWriteF(client->conn, "2001 %s ", sqlite3_column_text(stmt, 0))) != -1) {
                value = sqlite3_column_text(stmt, 1);
                ccode = ConnWriteF(client->conn, "%d\r\n%s", strlen(value), value);
                if (ccode != -1) {
                    continue;
                }
            }

            *failure = TRUE;
            break;
        }
    } else {
        *failure = TRUE;
        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
    }

    if (stmt) {
        sqlite3_reset(stmt);
        stmt = NULL;
    }

    return(ccode);
}

int 
NmapCommandDlist(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *guid;
    BOOL failure;
    FILE *fh;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;
    NLockStruct *lock = NULL;

    /* DLIST *(<guid>) */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = strchr(client->buffer + 5, ' ');

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (ptr) {
        ccode = NmapSqlPrepare(handle, NMAP_DREAD_BY_GUID, &(client->dstore.sql.sel.dread));
        if (ccode == SQLITE_OK) {
            stmt = client->dstore.sql.sel.dread;

            ccode = DStoreTranBegin(client, TRUE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode != SQLITE_DONE) {
            return(ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    while (ccode != -1) {
        memset(&info, 0, sizeof(info));

        if (ptr) {
            guid = ++ptr;
            ptr = strchr(ptr, ' ');
        } else {
            break;
        }

        if (ptr) {
            *ptr = '\0';

            if ((ptr - guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                break;
            }
        } else if (strlen(guid) == NMAP_GUID_LENGTH) {
            ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
        } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
            ccode = SQLITE_DONE;
            continue;
        } else {
            break;
        }

        if (ccode == SQLITE_OK) {
            memcpy(info.guid, guid, NMAP_GUID_LENGTH);

            ccode = sqlite3_step(stmt);
            if (ccode == SQLITE_ROW) {
                info.type = sqlite3_column_int(stmt, 0);
                info.start.store = sqlite3_column_int(stmt, 1);
                info.start.doc = info.start.store + sqlite3_column_int(stmt, 2);
                info.length = sqlite3_column_int(stmt, 3);
                strcpy(info.resource, sqlite3_column_text(stmt, 4));

                if (ReadNLockAquire(&lock, &client->userHash, client->user, info.resource, 3000) == NLOCK_AQUIRED) {
                    info.indexed = sqlite3_column_int(stmt, 5);
                    strcpy(info.version.store, sqlite3_column_text(stmt, 6));

                    
                    if (NmapCheckAuthorization(client, info.guid, NMAP_PRIV_READ, 0)) {
                        failure = TRUE;
                    } else {
                        ccode = DStoreListProperties(client, &info, &failure);
                    }
                    if (!failure && (ccode != -1)) {
                        switch (info.type) {
                            case NMAP_DOCTYPE_UNKNOWN_FOLDER: 
                            case NMAP_DOCTYPE_MAIL_FOLDER: 
                            case NMAP_DOCTYPE_CAL_FOLDER: 
                            case NMAP_DOCTYPE_AB_FOLDER: {
                                ccode = DStoreListFolder(client, &info);
                                break;
                            }

                            case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
                            case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
                            case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
                            case NMAP_DOCTYPE_AB_SHARED_FOLDER: {
                                ccode = DStoreListFolder(client, &info);
                                break;
                            }

                            case NMAP_DOCTYPE_UNKNOWN: 
                            case NMAP_DOCTYPE_MAIL: 
                            case NMAP_DOCTYPE_CAL: 
                            case NMAP_DOCTYPE_AB: {
                                fh = fopen(client->path, "rb");
                                if (fh) {
                                    if ((ccode != -1) 
                                            && ((ccode = ConnWriteF(client->conn, "2002 %d bytes in document.\r\n", info.length)) != -1)) {
                                        fseek(fh, info.start.doc, SEEK_SET);
                                        ccode = ConnWriteFromFile(client->conn, fh, info.length);
                                        if (ccode != -1) {
                                            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
                                        }
                                    }

                                    fclose(fh);
                                    fh = NULL;
                                } else if (ccode != -1) {
                                    if (info.type == NMAP_DOCTYPE_UNKNOWN) {
                                        ccode = ConnWrite(client->conn, "4224 Can't read store.\r\n", 24);
                                    } else if (info.type == NMAP_DOCTYPE_CAL) {
                                        ccode = ConnWrite(client->conn, MSG4224CANTREADCAL, sizeof(MSG4224CANTREADCAL) - 1);
                                    } else {
                                        ccode = ConnWrite(client->conn, "4224 Can't read address book.\r\n", 31);
                                    }
                                }

                                break;
                            }

                            case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: 
                            case NMAP_DOCTYPE_MAIL_IN_SHARE: 
                            case NMAP_DOCTYPE_CAL_IN_SHARE: 
                            case NMAP_DOCTYPE_AB_IN_SHARE: 
                            case NMAP_DOCTYPE_PURGED: 
                            default: {
                                ccode = ConnWrite(client->conn, "5000 Command not implemented.\r\n", 31);
                                break;
                            }
                        }
                    }

                    ReadNLockRelease(lock);
                } else {
                    switch (info.type) {
                        case NMAP_DOCTYPE_UNKNOWN_FOLDER: 
                        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_UNKNOWN: 
                        case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: {
                            ccode = ConnWrite(client->conn, "4120 Store locked.\r\n", 20);
                            break;
                        }

                        case NMAP_DOCTYPE_MAIL_FOLDER: 
                        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_MAIL: 
                        case NMAP_DOCTYPE_MAIL_IN_SHARE: {
                            ccode = ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1);
                            break;
                        }

                        case NMAP_DOCTYPE_CAL_FOLDER: 
                        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_CAL: 
                        case NMAP_DOCTYPE_CAL_IN_SHARE: {
                            ccode = ConnWrite(client->conn, MSG4120CALLOCKED, sizeof(MSG4120CALLOCKED) - 1);
                            break;
                        }

                        case NMAP_DOCTYPE_AB_FOLDER: 
                        case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_AB: 
                        case NMAP_DOCTYPE_AB_IN_SHARE: {
                            ccode = ConnWrite(client->conn, "4120 Address book locked.\r\n", 20);
                            break;
                        }

                        case NMAP_DOCTYPE_PURGED: 
                        default: {
                            ccode = ConnWrite(client->conn, "4240 Not permitted.\r\n", 21);
                            break;

                        }
                    }
                }
            } else if (ccode == SQLITE_DONE) {
                ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        } else {
            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
        }

        sqlite3_reset(stmt);
    }

    DStoreTranEnd(client);

    return(ccode);
}

static int 
DStoreUnknown(NMAPClient *client, NmapDStoreInfo *info)
{
    int ccode;
    BOOL commit;
    FILE *fh;
    struct stat sb;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt = NULL;
    sqlite3_stmt *sver = NULL;
    NLockStruct *lock = NULL;

    if ((ReadNLockAquire(&lock, &client->userHash, client->user, info->resource, 3000) == NLOCK_AQUIRED) 
            && (WriteNLockAquire(lock, 2000) == NLOCK_AQUIRED)) {
        sprintf(client->path, "%s/%s/%s", client->store, client->user, info->resource);
        if (stat(client->path, &sb) == 0) {
            fh = fopen(client->path, "ab");
        } else {
            WriteNLockRelease(lock);
            ReadNLockRelease(lock);

            return(ConnWrite(client->conn, "4224 No store.\r\n", 16));
        }
    } else {
        if (lock) {
            ReadNLockRelease(lock);
        }

        return(ConnWrite(client->conn, "4120 Store locked.\r\n", 20));
    }

    if (fh) {
        info->start.store = sb.st_size;

        do {
            if (((info->start.doc = WriteDocumentHeaders(client, fh, info->guid, info->length)) != -1) 
                    && ((ccode = ConnWrite(client->conn, "2002 Send document.\r\n", 21)) != -1) 
                    && ((ccode = ConnFlush(client->conn)) != -1) 
                    && ((ccode = ConnReadToFile(client->conn, fh, info->length)) != -1) 
                    && (fwrite("\r\n\r\n", sizeof(unsigned char), 4, fh) == 4)) {
                ccode = DStoreTranBegin(client, TRUE);
                if (ccode == SQLITE_DONE) {
                    commit = TRUE;
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
                    break;
                }
            } else {
                commit = FALSE;
                ccode = ConnWrite(client->conn, "4228 Can't write to store.\r\n", 28);
                break;
            }

            if (info->version.store[0]) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_UPDATE, &(client->dstore.sql.upd.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.upd.dstor;

                    if (((ccode = sqlite3_bind_int(stmt, 1, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->indexed)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_text(stmt, 5, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            } else {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_INSERT, &(client->dstore.sql.ins.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.ins.dstor;

                    if ((ccode == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->type)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 5, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 6, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_int(stmt, 7, info->indexed);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            }

            if (ccode == SQLITE_OK) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK) {
                    sver = client->dstore.sql.sel.dstor;

                    ccode = sqlite3_bind_text(sver, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                }
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);

                sqlite3_reset(stmt);
                stmt = NULL;

                if (ccode == SQLITE_DONE) {
                    ccode = sqlite3_step(sver);
                    if (ccode == SQLITE_ROW) {
                        strcpy(info->version.store, sqlite3_column_text(sver, 0));

                        sqlite3_reset(sver);
                        sver = NULL;

                        ccode = DStoreTranEnd(client);
                        if (ccode == SQLITE_DONE) {
                            ccode = ConnWriteF(client->conn, "1000 %s %s\r\n", info->guid, info->version.store);
                            break;
                        }
                    }
                }

                commit = FALSE;
            }

            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);

            DStoreTranAbort(client);
            break;
        } while (TRUE);

        if (stmt) {
            sqlite3_reset(stmt);
            stmt = NULL;
        }

        if (sver) {
            sqlite3_reset(sver);
            sver = NULL;
        }

        if (!commit) {
            fseek(fh, info->start.store, SEEK_SET);
            fwrite("", sizeof(unsigned char), 1, fh);
        }

        if (!(NMAP.flushFlags & FLUSH_BOX_ON_STORE)) {
            fclose(fh);
        } else {
            XplFlushWrites(fh);
            fclose(fh);
        }

        fh = NULL;
    } else if (errno == ENOSPC) {
        ccode = ConnWrite(client->conn, MSG5220QUOTAERR, sizeof(MSG5220QUOTAERR) - 1);
    } else {
        ccode = ConnWrite(client->conn, "4228 Can't write to store.\r\n", 28);
    }

    WriteNLockRelease(lock);
    ReadNLockRelease(lock);
    NotifyIndexer(client->user);

    return(ccode);
}

static int 
DStoreMail(NMAPClient *client, NmapDStoreInfo *info)
{
    int ccode;
    BOOL commit = FALSE;
    FILE *fh;
    struct stat sb;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt = NULL;
    sqlite3_stmt *sver = NULL;

    if ((ReadNLockAquire(&client->mail.lock, &client->userHash, client->user, info->resource, 3000) == NLOCK_AQUIRED) 
            && (WriteNLockAquire(client->mail.lock, 2000) == NLOCK_AQUIRED)) {
        sprintf(client->path, "%s/%s/%s.box", client->store, client->user, info->resource);
        if (stat(client->path, &sb) == 0) {
            fh = fopen(client->path, "ab");
        } else {
            WriteNLockRelease(client->mail.lock);
            ReadNLockRelease(client->mail.lock);
            client->mail.lock = NULL;

            return(ConnWrite(client->conn, MSG4224NOMBOX, sizeof(MSG4224NOMBOX) - 1));
        }
    } else {
        if (client->mail.lock) {
            ReadNLockRelease(client->mail.lock);
            client->mail.lock = NULL;
        }

        return(ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1));
    }

    if (fh) {
        info->start.store = sb.st_size;

        do {
            if (((info->start.doc = WriteDocumentHeaders(client, fh, info->guid, info->length)) != -1) 
                    && ((ccode = ConnWrite(client->conn, "2002 Send document.\r\n", 21)) != -1) 
                    && ((ccode = ConnFlush(client->conn)) != -1) 
                    && ((ccode = ConnReadToFile(client->conn, fh, info->length)) != -1) 
                    && (fwrite("\r\n\r\n", sizeof(unsigned char), 4, fh) == 4)) {
                ccode = DStoreTranBegin(client, TRUE);
                if (ccode != SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
                    break;
                }
            } else {
                ccode = ConnWrite(client->conn, MSG4228CANTWRITEMBOX, sizeof(MSG4228CANTWRITEMBOX) - 1);
                break;
            }

            if (info->version.store[0]) {
                /*
                    DMSMITH - fixme
                */
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_UPDATE, &(client->dstore.sql.upd.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.upd.dstor;

                    if (((ccode = sqlite3_bind_int(stmt, 1, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->indexed)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_text(stmt, 5, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                    }
                }
            } else {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_INSERT, &(client->dstore.sql.ins.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.ins.dstor;

                    if ((ccode == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->type)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 5, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 6, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_int(stmt, 7, info->indexed);
                    }
                }
            }

            if (ccode == SQLITE_OK) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK) {
                    sver = client->dstore.sql.sel.dstor;

                    ccode = sqlite3_bind_text(sver, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                }
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);

                sqlite3_reset(stmt);
                stmt = NULL;

                if (ccode == SQLITE_DONE) {
                    ccode = sqlite3_step(sver);
                    if (ccode == SQLITE_ROW) {
                        strcpy(info->version.store, sqlite3_column_text(sver, 0));

                        sqlite3_reset(sver);
                        sver = NULL;

                        ccode = DStoreTranEnd(client);
                        if (ccode == SQLITE_DONE) {
                            commit = TRUE;
                            break;
                        }
                    }
                }
            }

            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);

            DStoreTranAbort(client);
            break;
        } while (TRUE);
        
        if (stmt) {
            sqlite3_reset(stmt);
            stmt = NULL;
        }

        if (sver) {
            sqlite3_reset(sver);
            sver = NULL;
        }

        if (!commit) {
            fseek(fh, info->start.store, SEEK_SET);
            fwrite("", sizeof(unsigned char), 1, fh);
        }

        if (!(NMAP.flushFlags & FLUSH_BOX_ON_STORE)) {
            fclose(fh);
        } else {
            XplFlushWrites(fh);
            fclose(fh);
        }

        fh = NULL;
    } else if (errno == ENOSPC) {
        ccode = ConnWrite(client->conn, MSG5220QUOTAERR, sizeof(MSG5220QUOTAERR) - 1);
    } else {
        ccode = ConnWrite(client->conn, MSG4228CANTWRITEMBOX, sizeof(MSG4228CANTWRITEMBOX) - 1);
    }

    WriteNLockRelease(client->mail.lock);

    if (commit) {
        strcpy(client->mail.name, info->resource);

        ParseMailStore(client);

        client->mail.name[0] = '\0';
        client->mail.info.count = 0;
        client->mail.info.used = 0;

        ccode = ConnWriteF(client->conn, "1000 %s %s\r\n", info->guid, info->version.store);
    }

    ReadNLockRelease(client->mail.lock);
    client->mail.lock = NULL;
    NotifyIndexer(client->user);

    return(ccode);
}

static int 
DStoreCalendar(NMAPClient *client, NmapDStoreInfo *info)
{
    int ccode;
    BOOL commit;
    FILE *fh;
    struct stat sb;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt = NULL;
    sqlite3_stmt *sver = NULL;

    if ((ReadNLockAquire(&client->mail.lock, &client->userHash, client->user, info->resource, 3000) == NLOCK_AQUIRED) 
            && (WriteNLockAquire(client->mail.lock, 2000) == NLOCK_AQUIRED)) {
        sprintf(client->path, "%s/%s/%s"NMAP_CAL_EXT, client->store, client->user, info->resource);
        if (stat(client->path, &sb) == 0) {
            fh = fopen(client->path, "ab");
        } else {
            WriteNLockRelease(client->mail.lock);
            ReadNLockRelease(client->mail.lock);
            client->mail.lock = NULL;

            return(ConnWrite(client->conn, MSG4224NOCSTORE, sizeof(MSG4224NOCSTORE) - 1));
        }
    } else {
        if (client->mail.lock) {
            ReadNLockRelease(client->mail.lock);
            client->mail.lock = NULL;
        }

        return(ConnWrite(client->conn, MSG4120CALLOCKED, sizeof(MSG4120CALLOCKED) - 1));
    }

    if (fh) {
        info->start.store = sb.st_size;

        do {
            if (((info->start.doc = WriteDocumentHeaders(client, fh, info->guid, info->length)) != -1) 
                    && ((ccode = ConnWrite(client->conn, "2002 Send document.\r\n", 21)) != -1) 
                    && ((ccode = ConnFlush(client->conn)) != -1) 
                    && ((ccode = ConnReadToFile(client->conn, fh, info->length)) != -1) 
                    && (fwrite("\r\n\r\n", sizeof(unsigned char), 4, fh) == 4)) {
                ccode = DStoreTranBegin(client, TRUE);
                if (ccode == SQLITE_DONE) {
                    commit = TRUE;
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
                    break;
                }
            } else {
                commit = FALSE;
                ccode = ConnWrite(client->conn, MSG4228CANTWRITECAL, sizeof(MSG4228CANTWRITECAL) - 1);
                break;
            }

            if (info->version.store[0]) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_UPDATE, &(client->dstore.sql.upd.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.upd.dstor;

                    if (((ccode = sqlite3_bind_int(stmt, 1, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->indexed)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_text(stmt, 5, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            } else {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_INSERT, &(client->dstore.sql.ins.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.ins.dstor;

                    if ((ccode == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->type)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 5, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 6, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_int(stmt, 7, info->indexed);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            }

            if (ccode == SQLITE_OK) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK) {
                    sver = client->dstore.sql.sel.dstor;

                    ccode = sqlite3_bind_text(sver, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                }
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);

                sqlite3_reset(stmt);
                stmt = NULL;

                if (ccode == SQLITE_DONE) {
                    ccode = sqlite3_step(sver);
                    if (ccode == SQLITE_ROW) {
                        strcpy(info->version.store, sqlite3_column_text(sver, 0));

                        sqlite3_reset(sver);
                        sver = NULL;

                        ccode = DStoreTranEnd(client);
                        if (ccode == SQLITE_DONE) {
                            ccode = ConnWriteF(client->conn, "1000 %s %s\r\n", info->guid, info->version.store);
                            break;
                        }
                    }
                }

                commit = FALSE;
            }

            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);

            DStoreTranAbort(client);
            break;
        } while (TRUE);

        if (stmt) {
            sqlite3_reset(stmt);
            stmt = NULL;
        }

        if (sver) {
            sqlite3_reset(sver);
            sver = NULL;
        }

        if (!commit) {
            fseek(fh, info->start.store, SEEK_SET);
            fwrite("", sizeof(unsigned char), 1, fh);
        }

        if (!(NMAP.flushFlags & FLUSH_BOX_ON_STORE)) {
            fclose(fh);
        } else {
            XplFlushWrites(fh);
            fclose(fh);
        }

        fh = NULL;
    } else if (errno == ENOSPC) {
        ccode = ConnWrite(client->conn, MSG5220QUOTAERR, sizeof(MSG5220QUOTAERR) - 1);
    } else {
        ccode = ConnWrite(client->conn, MSG4228CANTWRITECAL, sizeof(MSG4228CANTWRITECAL) - 1);
    }

    WriteNLockRelease(client->mail.lock);
    ReadNLockRelease(client->mail.lock);
    client->mail.lock = NULL;
    NotifyIndexer(client->user);

    return(ccode);
}

static int 
DStoreAddressBook(NMAPClient *client, NmapDStoreInfo *info)
{
    int ccode;
    BOOL commit;
    FILE *fh;
    struct stat sb;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt = NULL;
    sqlite3_stmt *sver = NULL;
    NLockStruct *lock = NULL;

    if ((ReadNLockAquire(&lock, &client->userHash, client->user, info->resource, 3000) == NLOCK_AQUIRED) 
            && (WriteNLockAquire(lock, 2000) == NLOCK_AQUIRED)) {
        sprintf(client->path, "%s/%s/%s"NMAP_ADBK_EXT, client->store, client->user, info->resource);
        if (stat(client->path, &sb) == 0) {
            fh = fopen(client->path, "ab");
        } else {
            WriteNLockRelease(lock);
            ReadNLockRelease(lock);

            return(ConnWrite(client->conn, "4224 No address book.\r\n", 23));
        }
    } else {
        if (lock) {
            ReadNLockRelease(lock);
        }

        return(ConnWrite(client->conn, "4120 Address book locked.\r\n", 27));
    }

    if (fh) {
        info->start.store = sb.st_size;

        do {
            if (((info->start.doc = WriteDocumentHeaders(client, fh, info->guid, info->length)) != -1) 
                    && ((ccode = ConnWrite(client->conn, "2002 Send document.\r\n", 21)) != -1) 
                    && ((ccode = ConnFlush(client->conn)) != -1) 
                    && ((ccode = ConnReadToFile(client->conn, fh, info->length)) != -1) 
                    && (fwrite("\r\n\r\n", sizeof(unsigned char), 4, fh) == 4)) {
                ccode = DStoreTranBegin(client, TRUE);
                if (ccode == SQLITE_DONE) {
                    commit = TRUE;
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
                    break;
                }
            } else {
                commit = FALSE;
                ccode = ConnWrite(client->conn, "4228 Can't write to address book.\r\n", 28);
                break;
            }

            if (info->version.store[0]) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_UPDATE, &(client->dstore.sql.upd.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.upd.dstor;

                    if (((ccode = sqlite3_bind_int(stmt, 1, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->indexed)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_text(stmt, 5, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            } else {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_INSERT, &(client->dstore.sql.ins.dstor))) == SQLITE_OK) {
                    stmt = client->dstore.sql.ins.dstor;

                    if ((ccode == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 2, info->type)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 3, info->start.store)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 4, info->start.doc)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_int(stmt, 5, info->length)) == SQLITE_OK) 
                            && ((ccode = sqlite3_bind_text(stmt, 6, info->resource, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                        ccode = sqlite3_bind_int(stmt, 7, info->indexed);
                    } else {
                        commit = FALSE;
                    }
                } else {
                    commit = FALSE;
                }
            }

            if (ccode == SQLITE_OK) {
                if ((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK) {
                    sver = client->dstore.sql.sel.dstor;

                    ccode = sqlite3_bind_text(sver, 1, info->guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                }
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);

                sqlite3_reset(stmt);
                stmt = NULL;

                if (ccode == SQLITE_DONE) {
                    ccode = sqlite3_step(sver);
                    if (ccode == SQLITE_ROW) {
                        strcpy(info->version.store, sqlite3_column_text(sver, 0));

                        sqlite3_reset(sver);
                        sver = NULL;

                        ccode = DStoreTranEnd(client);
                        if (ccode == SQLITE_DONE) {
                            ccode = ConnWriteF(client->conn, "1000 %s %s\r\n", info->guid, info->version.store);
                            break;
                        }
                    }
                }

                commit = FALSE;
            }

            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);

            DStoreTranAbort(client);
            break;
        } while (TRUE);

        if (stmt) {
            sqlite3_reset(stmt);
            stmt = NULL;
        }

        if (sver) {
            sqlite3_reset(sver);
            sver = NULL;
        }

        if (!commit) {
            fseek(fh, info->start.store, SEEK_SET);
            fwrite("", sizeof(unsigned char), 1, fh);
        }

        if (!(NMAP.flushFlags & FLUSH_BOX_ON_STORE)) {
            fclose(fh);
        } else {
            XplFlushWrites(fh);
            fclose(fh);
        }

        fh = NULL;
    } else if (errno == ENOSPC) {
        ccode = ConnWrite(client->conn, MSG5220QUOTAERR, sizeof(MSG5220QUOTAERR) - 1);
    } else {
        ccode = ConnWrite(client->conn, "4228 Can't write to address book.\r\n", 28);
    }

    WriteNLockRelease(lock);
    ReadNLockRelease(lock);
    NotifyIndexer(client->user);

    return(ccode);
}

int 
NmapCommandDmove(void *param)
{
    int ccode;
    NMAPClient *client = (NMAPClient *)param;

    ccode = ConnWrite(client->conn, MSG1000NOTIMP, sizeof(MSG1000NOTIMP) - 1);

    /* will require the following permissions: 
       - unbind on original dir 
       - bind on target dir
       - if target exists, unbind on target dir
    */

    return(ccode);
}

int 
NmapCommandDpget(void *param)
{
    int ccode;
    unsigned char *guid;
    unsigned char *name;
    BOOL failure;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;

    /* DPGET [<guid> [*(<property name>)]] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            guid = client->buffer + 5;

            memset(&info, 0, sizeof(info));

            handle = client->dstore.handle;
        } else if (client->states & NMAP_CLIENT_MBOX) {
            return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
        } else {
            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if ((*guid++ == ' ') && (*guid != ' ')) {
        name = strchr(guid, ' ');
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    
    if (!name) {
        if (strlen(guid) == NMAP_GUID_LENGTH) {
            memcpy(&info.guid, guid, NMAP_GUID_LENGTH);
        } else {
            return(ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1));
        }

        if (((ccode = NmapSqlPrepare(handle, NMAP_DREAD_BY_GUID, &(client->dstore.sql.sel.dread))) == SQLITE_OK)
                && ((ccode = NmapSqlPrepare(handle, NMAP_DPGET_SELECT, &(client->dstore.sql.sel.dpget))) == SQLITE_OK)) {
            stmt = client->dstore.sql.sel.dread;

            ccode = DStoreTranBegin(client, TRUE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode == SQLITE_DONE) {
            if (!NmapCheckAuthorization(client, info.guid, NMAP_PRIV_READ, 0)) {
                if (((ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK) 
                    && ((ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK)) {
                    stmt = client->dstore.sql.sel.dread;
                }
                
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    info.type = sqlite3_column_int(stmt, 0);
                    info.start.store = sqlite3_column_int(stmt, 1);
                    info.start.doc = info.start.store + sqlite3_column_int(stmt, 2);
                    info.length = sqlite3_column_int(stmt, 3);
                    strcpy(info.resource, sqlite3_column_text(stmt, 4));
                    info.indexed = sqlite3_column_int(stmt, 5);
                    strcpy(info.version.store, sqlite3_column_text(stmt, 6));
                    
                    ccode = DStoreListProperties(client, &info, &failure);
                    if ((ccode != -1) && !failure) {
                        ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
                    }
                } else if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                } else {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
                sqlite3_reset(stmt);
            }
        } else {
            ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
        }
        
        DStoreTranEnd(client);

        return(ccode);
    }

    return(ConnWrite(client->conn, "5000 Command not implemented.\r\n", 31));
}

int 
NmapCommandDpset(void *param)
{
    int ccode;
    unsigned long length;
    unsigned char *ptr;
    unsigned char *name;
    unsigned char *value;
    unsigned char guid[NMAP_GUID_LENGTH + 1];
    BOOL commit;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;

    /* DPSET <guid> <property name> <property length> */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            name = client->buffer + 5;

            handle = client->dstore.handle;
        } else if (client->states & NMAP_CLIENT_MBOX) {
            return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
        } else {
            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if ((*name++ == ' ') 
            && (*name != ' ') 
            && ((name = strchr(name, ' ')) != NULL) 
            && ((++name - client->buffer) == (NMAP_GUID_LENGTH + 7)) 
            && (*name != ' ') 
            && ((ptr = strchr(name, ' ')) != NULL) 
            && isdigit(ptr[1])) {
        name[-1] = '\0';
        *ptr++ = '\0';

        if (!strchr(name, '.') || 
            (!XplStrNCaseCmp(name, "nmap.", 5) && strcmp(name, "nmap.access-control")))
        {
            return(ConnWrite(client->conn, "3012 Restricted property name.\r\n", 32));
        } 

        memcpy(guid, client->buffer + 6, NMAP_GUID_LENGTH + 1);
        length = atol(ptr);
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (length) {
        value = (unsigned char *)MemMalloc(length + 1);
        if (value 
                && ((ccode = ConnWriteF(client->conn, "2002 Send value.\r\n", 18)) != -1) 
                && ((ccode = ConnFlush(client->conn)) != -1) 
                && ((ccode = ConnReadCount(client->conn, value, length)) != -1)) {
            value[length] = '\0';
        } else if (!value) {
            return(ConnWrite(client->conn, MSG5001NOMEMORY, sizeof(MSG5001NOMEMORY) - 1));
        } else {
            return(ccode);
        }
    } else {
        value = NULL;
    }

    /* fixme: if name is "nmap.access-control", then validate ACL */

    if (((ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_DPSET_SELECT, &(client->dstore.sql.sel.dpset))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_DPSET_INSERT, &(client->dstore.sql.ins.dpset))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_DPSET_UPDATE, &(client->dstore.sql.upd.dpset))) == SQLITE_OK) 
            && ((ccode = NmapSqlPrepare(handle, NMAP_DPSET_DELETE, &(client->dstore.sql.del.dpset))) == SQLITE_OK)
            && ((ccode = NmapSqlPrepare(handle, NMAP_DINSET_UPDATE, &(client->dstore.sql.upd.dinset))) == SQLITE_OK)) {
        ccode = DStoreTranBegin(client, TRUE);
        if (ccode == SQLITE_DONE) {
            stmt = client->dstore.sql.sel.dstor;
            commit = TRUE;
        } else {
            return(ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if ((ccode = sqlite3_bind_text(stmt, 1, guid, -1, SQLITE_STATIC)) == SQLITE_OK) {
        ccode = sqlite3_step(stmt);
    } else {
        commit = FALSE;
        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
    }

    sqlite3_reset(stmt);

    if (commit) {
        if (ccode == SQLITE_ROW) {
            stmt = client->dstore.sql.sel.dpset;
            if (((ccode = sqlite3_bind_text(stmt, 1, guid, -1, SQLITE_STATIC)) == SQLITE_OK) 
                    && ((ccode = sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                ccode = sqlite3_step(stmt);
            } else {
                commit = FALSE;
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);
        } else {
            commit = FALSE;
            if (ccode == SQLITE_DONE) {
                ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        }
    }

    if (commit) {
        /* Unset the indexed flag */
        stmt = client->dstore.sql.upd.dinset;
        if (((ccode = sqlite3_bind_int(stmt, 1, 0)) == SQLITE_OK)
            && ((ccode = sqlite3_bind_text(stmt, 2, guid, NMAP_GUID_LENGTH, SQLITE_STATIC)) == SQLITE_OK)) {
            ccode = sqlite3_step(stmt);
        } else {
            commit = FALSE;
            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
        }
        
        sqlite3_reset(stmt);
    }
    
    if (commit) {
        if (NmapCheckAuthorization (client, guid, NMAP_PRIV_WRITE_PROPS, 0)) {
            commit = FALSE;
        }
    }

    if (commit) {
        if ((ccode == SQLITE_DONE) || (ccode == SQLITE_ROW)) {
            if (!length) {
                ccode = SQLITE_NOTFOUND;
            }
        } else {
            ccode = SQLITE_ERROR;
        }

        switch (ccode) {
            case SQLITE_NOTFOUND: {
                stmt = client->dstore.sql.del.dpset;
                if (((ccode = sqlite3_bind_text(stmt, 1, guid, -1, SQLITE_STATIC)) == SQLITE_OK) 
                        && ((ccode = sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                    ccode = sqlite3_step(stmt);
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }

                sqlite3_reset(stmt);
                break;
            }

            case SQLITE_DONE: {
                stmt = client->dstore.sql.ins.dpset;
                if (((ccode = sqlite3_bind_text(stmt, 1, guid, -1, SQLITE_STATIC)) == SQLITE_OK) 
                        && ((ccode = sqlite3_bind_text(stmt, 2, name, -1, SQLITE_STATIC)) == SQLITE_OK) 
                        && ((ccode = sqlite3_bind_text(stmt, 3, value, length, SQLITE_STATIC)) == SQLITE_OK)) {
                    ccode = sqlite3_step(stmt);
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }

                sqlite3_reset(stmt);
                break;
            }

            case SQLITE_ROW: {
                stmt = client->dstore.sql.upd.dpset;
                if (((ccode = sqlite3_bind_text(stmt, 1, value, -1, SQLITE_STATIC)) == SQLITE_OK) 
                        && ((ccode = sqlite3_bind_text(stmt, 2, guid, -1, SQLITE_STATIC)) == SQLITE_OK) 
                        && ((ccode = sqlite3_bind_text(stmt, 3, name, -1, SQLITE_STATIC)) == SQLITE_OK)) {
                    ccode = sqlite3_step(stmt);
                } else {
                    commit = FALSE;
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }

                sqlite3_reset(stmt);
                break;
            }

            case SQLITE_ERROR:
            default: {
                commit = FALSE;
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                break;
            }
        }
    }

    if (value) {
        MemFree(value);
    }

    if (commit) {
        if (ccode == SQLITE_DONE) {
            ccode = DStoreTranEnd(client);
        }

        if (ccode == SQLITE_DONE) {
            return(ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1));
        }

        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
    }

    DStoreTranAbort(client);

    return(ccode);
}

int 
NmapCommandDpurg(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *guid;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    sqlite3_stmt *dpstmt;  /* for the document properties */
    sqlite3_stmt *dstmt;   /* for the document itself */
    sqlite3_stmt *distmt;  /* for inserting into the purged table */

    /* DPURG [*(<NMAP GUID>)] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 5;

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (*ptr == ' ') {
        ccode = NmapSqlPrepare(handle, NMAP_DINFO_BY_GUID, &(client->dstore.sql.sel.dinfoGuid)) ||
            NmapSqlPrepare(handle, NMAP_DPURG_BY_GUID, &(client->dstore.sql.sel.dpurgGuid)) ||
            NmapSqlPrepare(handle, NMAP_DPURG_PROPS_BY_GUID, &(client->dstore.sql.sel.dpurgPropsGuid)) ||
            NmapSqlPrepare(handle, NMAP_DPURG_GUID_INSERT, &(client->dstore.sql.sel.dpurgGuidInsert));
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    stmt = client->dstore.sql.sel.dinfoGuid;
    dstmt = client->dstore.sql.sel.dpurgGuid;
    dpstmt = client->dstore.sql.sel.dpurgPropsGuid;
    distmt = client->dstore.sql.sel.dpurgGuidInsert;

    if (ccode == SQLITE_OK) {
        ccode = DStoreTranBegin(client, TRUE);
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (ccode == SQLITE_DONE) {
        while (ccode != -1) {
            if (ptr) {
                guid = ++ptr;
                ptr = strchr(ptr, ' ');
            } else {
                break;
            }

            if (ptr) {
                *ptr = '\0';

                if ((ptr - guid) == NMAP_GUID_LENGTH) {
                    ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                    ccode = SQLITE_DONE;
                    continue;
                } else {
                    break;
                }
            } else if (strlen(guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                break;
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    int type;
                    char containerGuid[NMAP_GUID_LENGTH+1];

                    type = sqlite3_column_int(stmt, 1);

                    if (type & NMAP_DOCTYPE_FOLDER) {
                        ConnWrite(client->conn, "5000 Folder removal not implemented.\r\n", 39);
                    } else {
                        ccode =
                            FindResourceGuid(client, sqlite3_column_text(stmt, 0), type | NMAP_DOCTYPE_FOLDER, containerGuid) ||
                            NmapCheckAuthorization(client, containerGuid, NMAP_PRIV_UNBIND, 0) ||
                            NmapCheckAuthorization(client, guid, NMAP_PRIV_WRITE, 0);
                        sqlite3_reset(stmt);

                        if (!ccode) {
                            ccode = sqlite3_bind_text(dpstmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                            if (ccode == SQLITE_OK) {
                                ccode = sqlite3_step(dpstmt);
                                sqlite3_reset(dpstmt);

                                ccode = sqlite3_bind_text(dstmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                                ccode = sqlite3_step(dstmt);
                                sqlite3_reset(dstmt);
                            }

                            if (ccode != SQLITE_DONE) {
                                ccode = ConnWrite(client->conn, "1001 Not Deleted\r\n", 18);
                            }

                            ccode = sqlite3_bind_text(distmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);

                            if (ccode == SQLITE_OK) {
                                ccode = sqlite3_step(distmt);
                                sqlite3_reset(distmt);
                            }

                            ccode = ConnWrite(client->conn, "1000 " MSG1000DELETED "\r\n", sizeof(MSG1000DELETED) + 6);
                        } else {
                            ccode = ConnWrite(client->conn, MSG4240NOPERMISSION, sizeof(MSG4240NOPERMISSION) - 1);
                        }
                    }
                } else if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                } else {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        }

        DStoreTranEnd(client);

        if (ccode != -1) {
            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
        }
    } else {
        ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
    }

    NotifyIndexer(client->user);

    return(ccode);
}

int 
NmapCommandDlspg(void *param)
{
    int ccode;
    unsigned char *ptr;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt;

    /* DLSPG */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 5;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        ccode = NmapSqlPrepare(handle, NMAP_DLSPG_SELECT, &(client->dstore.sql.sel.dlspg));
        if (ccode == SQLITE_OK) {
            ccode = DStoreTranBegin(client, FALSE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode == SQLITE_DONE) {
            stmt = client->dstore.sql.sel.dlspg;

            while ((ccode = sqlite3_step(stmt)) == SQLITE_ROW) {
                if ((ccode = ConnWriteF(client->conn, "2001 %s\r\n", 
                                        sqlite3_column_text(stmt, 0))) != -1) {
                    continue;
                }

                break;
            }

            if (ccode == SQLITE_DONE) {
                ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);

            DStoreTranEnd(client);
        } else {
            ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
        }

        return(ccode);
    }

    return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
}

int 
NmapCommandDrmpg(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *guid;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle;
    sqlite3_stmt *stmt;

    /* DRMPG [*(<NMAP GUID>)] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 5;

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (*ptr == ' ') {
        ccode = NmapSqlPrepare(handle, NMAP_DRMPG_GUID, &(client->dstore.sql.sel.drmpg));
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (ccode == SQLITE_OK) {
        ccode = DStoreTranBegin(client, TRUE);
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (ccode == SQLITE_DONE) {
        stmt = client->dstore.sql.sel.drmpg;

        while (ccode != -1) {
            if (ptr) {
                guid = ++ptr;
                ptr = strchr(ptr, ' ');
            } else {
                break;
            }

            if (ptr) {
                *ptr = '\0';

                if ((ptr - guid) == NMAP_GUID_LENGTH) {
                    ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                    ccode = SQLITE_DONE;
                    continue;
                } else {
                    break;
                }
            } else if (strlen(guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                break;
            }

            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                sqlite3_reset(stmt);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        }

        DStoreTranEnd(client);

        if (ccode != -1) {
            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
        }
    } else {
        ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
    }

    return(ccode);
}

int 
NmapCommandDread(void *param)
{
    int ccode;
    unsigned char *ptr;
    unsigned char *guid;
    const unsigned char *text;
    FILE *fh;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;
    NLockStruct *lock = NULL;

    /* DREAD *(<guid>) */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = strchr(client->buffer + 5, ' ');

            handle = client->dstore.handle;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (ptr) {
        ccode = NmapSqlPrepare(handle, NMAP_DREAD_BY_GUID, &(client->dstore.sql.sel.dread));
        if (ccode == SQLITE_OK) {
            stmt = client->dstore.sql.sel.dread;

            ccode = DStoreTranBegin(client, TRUE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (ccode == SQLITE_DONE) {
        while (ccode != -1) {
            memset(&info, 0, sizeof(info));

            if (ptr) {
                guid = ++ptr;
                ptr = strchr(ptr, ' ');
            } else {
                break;
            }

            if (ptr) {
                *ptr = '\0';

                if ((ptr - guid) == NMAP_GUID_LENGTH) {
                    ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                    ccode = SQLITE_DONE;
                    continue;
                } else {
                    break;
                }
            } else if (strlen(guid) == NMAP_GUID_LENGTH) {
                ccode = sqlite3_bind_text(stmt, 1, guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            } else if ((ccode = ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1)) != -1) {
                ccode = SQLITE_DONE;
                continue;
            } else {
                break;
            }

            if (ccode == SQLITE_OK) {
                strcpy(info.guid, guid);

                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    info.type = sqlite3_column_int(stmt, 0);
                    info.start.store = sqlite3_column_int(stmt, 1);
                    info.start.doc = info.start.store + sqlite3_column_int(stmt, 2);
                    info.length = sqlite3_column_int(stmt, 3);
                    strcpy(info.resource, sqlite3_column_text(stmt, 4));

                    if (ReadNLockAquire(&lock, &client->userHash, client->user, info.resource, 3000) == NLOCK_AQUIRED) {
                        info.indexed = sqlite3_column_int(stmt, 5);
                        strcpy(info.version.store, sqlite3_column_text(stmt, 6));
                    } else {
                        lock = NULL;
                    }

                    if (NmapCheckAuthorization(client, info.guid, NMAP_PRIV_READ, 0)) {
                        if (lock) {
                            ReadNLockRelease(lock);
                            lock = NULL;
                        } 
                        sqlite3_reset(stmt);
                        continue;
                    }

                    switch (info.type) {
                        case NMAP_DOCTYPE_UNKNOWN_FOLDER: 
                        case NMAP_DOCTYPE_MAIL_FOLDER: 
                        case NMAP_DOCTYPE_CAL_FOLDER: 
                        case NMAP_DOCTYPE_AB_FOLDER: {
                            if (lock) {
                                ccode = DStoreListFolder(client, &info);
                            } else if (info.type == NMAP_DOCTYPE_UNKNOWN_FOLDER) {
                                ccode = ConnWrite(client->conn, "4120 Store locked.\r\n", 20);
                            } else if (info.type == NMAP_DOCTYPE_MAIL_FOLDER) {
                                ccode = ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1);
                            } else if (info.type == NMAP_DOCTYPE_CAL_FOLDER) {
                                ccode = ConnWrite(client->conn, MSG4120CALLOCKED, sizeof(MSG4120CALLOCKED) - 1);
                            } else if (info.type == NMAP_DOCTYPE_AB_FOLDER) {
                                ccode = ConnWrite(client->conn, "4120 Address book locked.\r\n", 20);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
                        case NMAP_DOCTYPE_AB_SHARED_FOLDER: {
                            if (lock) {
                                ccode = DStoreListFolder(client, &info);
                            } else if (info.type == NMAP_DOCTYPE_UNKNOWN_FOLDER) {
                                ccode = ConnWrite(client->conn, "4120 Store locked.\r\n", 20);
                            } else if (info.type == NMAP_DOCTYPE_MAIL_FOLDER) {
                                ccode = ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1);
                            } else if (info.type == NMAP_DOCTYPE_CAL_FOLDER) {
                                ccode = ConnWrite(client->conn, MSG4120CALLOCKED, sizeof(MSG4120CALLOCKED) - 1);
                            } else if (info.type == NMAP_DOCTYPE_AB_FOLDER) {
                                ccode = ConnWrite(client->conn, "4120 Address book locked.\r\n", 20);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_UNKNOWN: {
                            if (lock) {
                                text = sqlite3_column_text(stmt, 4);
                                ccode = sprintf(client->path, "%s/%s/%s", client->store, client->user, text);
                                fh = fopen(client->path, "rb");
                                if (fh) {
                                    if ((ccode = ConnWriteF(client->conn, "2002 %d bytes in document.\r\n", info.length)) != -1) {
                                        fseek(fh, info.start.doc, SEEK_SET);
                                        ccode = ConnWriteFromFile(client->conn, fh, info.length);
                                        if (ccode != -1) {
                                            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) -  1);
                                        }
                                    }

                                    fclose(fh);
                                    fh = NULL;
                                } else {
                                    ccode = ConnWrite(client->conn, "4224 Can't read store.\r\n", 24);
                                }
                            } else {
                                ccode = ConnWrite(client->conn, "4120 Store locked.\r\n", 20);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_MAIL: {
                            if (lock) {
                                text = sqlite3_column_text(stmt, 4);
                                ccode = sprintf(client->path, "%s/%s/%s.box", client->store, client->user, text);
                                fh = fopen(client->path, "rb");
                                if (fh) {
                                    if ((ccode = ConnWriteF(client->conn, "2002 %d bytes in document.\r\n", info.length)) != -1) {
                                        fseek(fh, info.start.doc, SEEK_SET);
                                        ccode = ConnWriteFromFile(client->conn, fh, info.length);
                                        if (ccode != -1) {
                                            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
                                        }
                                    }

                                    fclose(fh);
                                    fh = NULL;
                                } else {
                                    ccode = ConnWrite(client->conn, MSG4224CANTREADMBOX, sizeof(MSG4224CANTREADMBOX) - 1);
                                }
                            } else {
                                ccode = ConnWrite(client->conn, MSG4120BOXLOCKED, sizeof(MSG4120BOXLOCKED) - 1);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_CAL: {
                            if (lock) {
                                text = sqlite3_column_text(stmt, 4);
                                ccode = sprintf(client->path, "%s/%s/%s"NMAP_CAL_EXT, client->store, client->user, text);
                                fh = fopen(client->path, "rb");
                                if (fh) {
                                    if ((ccode = ConnWriteF(client->conn, "2002 %d bytes in document.\r\n", info.length)) != -1) {
                                        fseek(fh, info.start.doc, SEEK_SET);
                                        ccode = ConnWriteFromFile(client->conn, fh, info.length);
                                        if (ccode != -1) {
                                            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) -  1);
                                        }
                                    }

                                    fclose(fh);
                                    fh = NULL;
                                } else {
                                    ccode = ConnWrite(client->conn, MSG4224CANTREADCAL, sizeof(MSG4224CANTREADCAL) - 1);
                                }
                            } else {
                                ccode = ConnWrite(client->conn, MSG4120CALLOCKED, sizeof(MSG4120CALLOCKED) - 1);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_AB: {
                            if (lock) {
                                text = sqlite3_column_text(stmt, 4);
                                ccode = sprintf(client->path, "%s/%s/%s"NMAP_ADBK_EXT, client->store, client->user, text);
                                fh = fopen(client->path, "rb");
                                if (fh) {
                                    if ((ccode = ConnWriteF(client->conn, "2002 %d bytes in document.\r\n", info.length)) != -1) {
                                        fseek(fh, info.start.doc, SEEK_SET);
                                        ccode = ConnWriteFromFile(client->conn, fh, info.length);
                                        if (ccode != -1) {
                                            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) -  1);
                                        }
                                    }

                                    fclose(fh);
                                    fh = NULL;
                                } else {
                                    ccode = ConnWrite(client->conn, "4224 Can't read address book.\r\n", 31);
                                }
                            } else {
                                ccode = ConnWrite(client->conn, "4120 Address book locked.\r\n", 20);
                            }

                            break;
                        }

                        case NMAP_DOCTYPE_UNKNOWN_IN_SHARE: 
                        case NMAP_DOCTYPE_MAIL_IN_SHARE: 
                        case NMAP_DOCTYPE_CAL_IN_SHARE: 
                        case NMAP_DOCTYPE_AB_IN_SHARE: {
                            ccode = ConnWrite(client->conn, MSG1000NOTIMP, sizeof(MSG1000NOTIMP) - 1);
                            break;
                        }

                        case NMAP_DOCTYPE_PURGED: 
                        default: {
                            ccode = ConnWrite(client->conn, "4220 Not permitted.\r\n", 21);
                            break;
                        }
                    }

                    if (lock) {
                        ReadNLockRelease(lock);
                    }
                } else if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                } else {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);
        }

        DStoreTranEnd(client);
    } else {
        ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
    }

    return(ccode);
}

int 
NmapCommandDrepl(void *param)
{
    int ccode;
    int count;
    unsigned char *cur;
    unsigned char *next;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;

    memset(&info, 0, sizeof(info));

    handle = client->dstore.handle;

    /* DREPL <guid> <revision> <length> [I<index>] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            if (!(NMAP.flags & NMAP_FLAG_DISK_SPACE_LOW)) {
                cur = client->buffer + 5;
            } else {
                return(ConnWrite(client->conn, MSG5221SPACELOW, sizeof(MSG5221SPACELOW) - 1));
            }
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }

        count = sprintf(client->path, "%s/%s/", client->store, client->user);
        if ((*cur++ == ' ') && *cur) {
            next = strchr(cur, ' ');
            if (next && ((next - cur) == NMAP_GUID_LENGTH)) {
                *next++ = '\0';
                memcpy(info.guid, cur, NMAP_GUID_LENGTH + 1);

                cur = next;
                next = strchr(cur, ' ');
            }

            if (next) {
                *next++ = '\0';
                strcpy(info.version.source, cur);

                cur = next;
                next = strchr(cur, ' ');
                if (next) {
                    *next++ = '\0';
                }

                info.length = atoi(cur);
            }

            if (next) {
                if ((toupper(next[0]) == 'I') && (next[1] != '0') && (isdigit(next[1]))) {
                    info.indexed = atoi(++next);
                } else {
                    info.guid[0] = '\0';
                }
            }
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (info.guid[0] && info.version.source[0] && info.length) {
        ccode = DStoreTranBegin(client, FALSE);
        if (ccode == SQLITE_DONE) {
            ccode = NmapSqlPrepare(handle, NMAP_DINFO_BY_GUID, &(client->dstore.sql.sel.dinfoGuid));
        } else {
            return(ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1));
        }

        if (ccode == SQLITE_OK) {
            stmt = client->dstore.sql.sel.dinfoGuid;
            ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
            if (ccode == SQLITE_OK) {
                ccode = sqlite3_step(stmt);
                if (ccode == SQLITE_ROW) {
                    strcpy(info.resource, sqlite3_column_text(stmt, 0));
                    info.type = sqlite3_column_int(stmt, 1);
                    strcpy(info.version.store, sqlite3_column_text(stmt, 6));
                } else {
                    info.guid[0] = '\0';
                    if (ccode == SQLITE_DONE) {
                        ccode = ConnWrite(client->conn, MSG4220NOGUID, sizeof(MSG4220NOGUID) - 1);
                    } else {
                        ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                    }
                }
            } else {
                info.guid[0] = '\0';
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        DStoreTranEnd(client);

        if (info.guid[0] == '\0') {
            return(ccode);
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (info.version.source[0] != '-' 
        && strcmp(info.version.source, info.version.store)) {
        return(ConnWriteF(client->conn, "4221 Current document revision is %s.\r\n", info.version.store));
    }

    {
        char containerGuid[NMAP_GUID_LENGTH+1];
        ccode = 
            FindResourceGuid(client, info.resource, info.type | NMAP_DOCTYPE_FOLDER, containerGuid) ||
            NmapCheckAuthorization(client, containerGuid, NMAP_PRIV_WRITE_CONTENT, 0);
        if (ccode) {
            return ccode;
        }
    }

    switch (info.type) {
        case NMAP_DOCTYPE_UNKNOWN: {
            ccode = DStoreUnknown(client, &info);
            break;
        }

        case NMAP_DOCTYPE_MAIL: {
            if (info.version.source[0] == '\0') {
                ccode = DStoreMail(client, &info);
            } else {
                ccode = ConnWrite(client->conn, "4240 Not permitted.\r\n", 21);
            }

            break;
        }

        case NMAP_DOCTYPE_CAL: {
            ccode = DStoreCalendar(client, &info);
            break;
        }

        case NMAP_DOCTYPE_AB: {
            ccode = DStoreAddressBook(client, &info);
            break;
        }

        case NMAP_DOCTYPE_UNKNOWN_FOLDER: 
        case NMAP_DOCTYPE_MAIL_FOLDER: 
        case NMAP_DOCTYPE_CAL_FOLDER: 
        case NMAP_DOCTYPE_AB_FOLDER: 
        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
        case NMAP_DOCTYPE_MAIL_IN_SHARE: 
        case NMAP_DOCTYPE_CAL_IN_SHARE: 
        case NMAP_DOCTYPE_AB_IN_SHARE: 
        default: {
            ccode = ConnWrite(client->conn, "4240 Not permitted.\r\n", 21);
            break;
        }
    }

    return(ccode);
}

int 
NmapCommandDrsrc(void *param)
{
    int ccode;
    int type = 0;
    unsigned char *ptr;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;

    /* DRSRC [(<Type>)] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            memset(&info, 0, sizeof(info));

            handle = client->dstore.handle;

            ptr = client->buffer + 5;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if ((*ptr == '\0') || ((*ptr++ == ' ') && isdigit(*ptr) 
                           && NMAP_IS_FOLDER((type = atoi(ptr)))))
    {
        ccode = NmapSqlPrepare(handle, NMAP_DSTORE_SELECT, &(client->dstore.sql.sel.dinfo));
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    if (ccode == SQLITE_OK) {
        ccode = DStoreTranBegin(client, FALSE);
    } else {
        return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
    }

    if (ccode == SQLITE_DONE) {
        stmt = client->dstore.sql.sel.dinfo;

        while ((ccode = sqlite3_step(stmt)) == SQLITE_ROW) {
            strcpy(info.guid, sqlite3_column_text(stmt, 0));
            strcpy(info.resource, sqlite3_column_text(stmt, 1));
            info.type = sqlite3_column_int(stmt, 2);

            if (NmapCheckAuthorization(client, info.guid, NMAP_PRIV_READ, 1)) {
                continue;
            }

            if ((type == info.type) || ((type == 0) && (NMAP_IS_FOLDER(info.type)))) {
                info.indexed = sqlite3_column_int(stmt, 3);
                strcpy(info.version.store, sqlite3_column_text(stmt, 4));

                if ((ccode = ConnWriteF(client->conn, "2001 %s %s %d %d %s\r\n", 
                        info.guid, 
                        info.resource, 
                        info.type, 
                        info.indexed, 
                        info.version.store)) != -1) {
                    continue;
                }
            } else {
                continue;
            }

            break;
        }

        if (ccode == SQLITE_DONE) {
            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
        } else {
            ccode = ConnWrite(client->conn, MSG5004INTERNALERR, sizeof(MSG5004INTERNALERR) - 1);
        }

        sqlite3_reset(stmt);

        DStoreTranEnd(client);
    } else {
        ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
    }

    return(ccode);
}

int 
NmapCommandDstor(void *param)
{
    int ccode;
    int count;
    unsigned char *cur;
    unsigned char *next;
    sqlite3 *handle;
    sqlite3_stmt *stmt;
    NMAPClient *client = (NMAPClient *)param;
    NmapDStoreInfo info;

    memset(&info, 0, sizeof(info));

    handle = client->dstore.handle;

    /* DSTOR <resource> <type> <length> [I<index>] [G<guid>] */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            if (!(NMAP.flags & NMAP_FLAG_DISK_SPACE_LOW)) {
                cur = client->buffer + 5;
            } else {
                return(ConnWrite(client->conn, MSG5221SPACELOW, sizeof(MSG5221SPACELOW) - 1));
            }
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }

        count = sprintf(client->path, "%s/%s/", client->store, client->user);
        if ((*cur++ == ' ') && *cur) {
            next = strchr(cur, ' ');
            if (next) {
                *next++ = '\0';

                memcpy(info.resource, client->buffer + 6, next - cur);

                cur = next;
                next = strchr(cur, ' ');
            }

            if (next) {
                *next++ = '\0';

                info.type = atol(cur);
                if (!NMAP_IS_FOLDER(info.type)) {
                    cur = next;
                    next = strchr(cur, ' ');
                    if (next) {
                        *next++ = '\0';
                    }

                    info.length = atol(cur);
                    if (!info.length) {
                        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
                    }
                } else {
                    return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
                    
                }
            } else {
                return ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1);
            }

            while (next) {
                cur = next;
                next = strchr(next, ' ');
                if (next) {
                    *next++ = ' ';
                }

                switch (toupper(*cur++)) {
                    case 'G': {
                        if ((next && ((next - cur - 1) == NMAP_GUID_LENGTH)) 
                                || (!next && (strlen(cur) == NMAP_GUID_LENGTH))) {
                            memcpy(info.guid, cur, NMAP_GUID_LENGTH);
                            info.guid[NMAP_GUID_LENGTH] = '\0';
                            continue;
                        }
                        return(ConnWrite(client->conn, MSG3011BADGUIDLEN, sizeof(MSG3011BADGUIDLEN) - 1));
                    }

                    case 'I': {
                        info.indexed = atoi(cur);
                        break;
                    }

                    default: {
                        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
                    }
                }
            }
        } else {
            return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (info.resource[0] && info.length) {
        if (info.guid[0] == '\0') {
            GuidAlloc(info.guid);
            info.guid[NMAP_GUID_LENGTH] = '\0';
        } else {
            ccode = DStoreTranBegin(client, FALSE);
            if (ccode == SQLITE_DONE) {
                ccode = NmapSqlPrepare(handle, NMAP_DINFO_VERSION, &(client->dstore.sql.sel.dstor));
            } else {
                return(ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1));
            }

            if (ccode == SQLITE_OK) {
                stmt = client->dstore.sql.sel.dstor;
                ccode = sqlite3_bind_text(stmt, 1, info.guid, NMAP_GUID_LENGTH, SQLITE_STATIC);
                if (ccode == SQLITE_OK) {
                    ccode = sqlite3_step(stmt);
                    if (ccode != SQLITE_DONE) {
                        if (ccode == SQLITE_ROW) {
                            ccode = ConnWriteF(client->conn, "4226 Guid %s already exists\r\n", info.guid);
                        } else {
                            ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                        }

                        info.guid[0] = '\0';
                    }
                } else {
                    info.guid[0] = '\0';
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }

                sqlite3_reset(stmt);
            } else {
                info.guid[0] = '\0';
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            DStoreTranEnd(client);

            if (info.guid[0] == '\0') {
                return(ccode);
            }
        }
    } else {
        return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
    }

    {
        char containerGuid[NMAP_GUID_LENGTH+1];
        ccode = FindResourceGuid(client, info.resource, 
                                 info.type | NMAP_DOCTYPE_FOLDER, containerGuid);
        if ((ccode && -4224 != ccode)
            || NmapCheckAuthorization(client, containerGuid, NMAP_PRIV_BIND, 1))
        {
            return ccode;
        }
    }
        
    switch (info.type) {
        case NMAP_DOCTYPE_UNKNOWN: {
            ccode = DStoreUnknown(client, &info);
            break;
        }

        case NMAP_DOCTYPE_MAIL: {
            ccode = DStoreMail(client, &info);
            break;
        }

        case NMAP_DOCTYPE_CAL: {
            ccode = DStoreCalendar(client, &info);
            break;
        }

        case NMAP_DOCTYPE_AB: {
            ccode = DStoreAddressBook(client, &info);
            break;
        }

        case NMAP_DOCTYPE_UNKNOWN_FOLDER: 
        case NMAP_DOCTYPE_MAIL_FOLDER: 
        case NMAP_DOCTYPE_CAL_FOLDER: 
        case NMAP_DOCTYPE_AB_FOLDER: 
        case NMAP_DOCTYPE_UNKNOWN_SHARED_FOLDER: 
        case NMAP_DOCTYPE_MAIL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_CAL_SHARED_FOLDER: 
        case NMAP_DOCTYPE_AB_SHARED_FOLDER: 
        case NMAP_DOCTYPE_MAIL_IN_SHARE: 
        case NMAP_DOCTYPE_CAL_IN_SHARE: 
        case NMAP_DOCTYPE_AB_IN_SHARE: 
        default: {
            ccode = ConnWrite(client->conn, "4240 Not permitted.\r\n", 21);
            break;
        }
    }

    return(ccode);
}

int 
NmapCommandDunfo(void *param)
{
    int ccode;
    unsigned char *ptr;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt;

    /* DUNFO */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 5;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    if (*ptr == '\0') {
        ccode = NmapSqlPrepare(handle, NMAP_DUNFO_SELECT, &(client->dstore.sql.sel.dunfo));
        if (ccode == SQLITE_OK) {
            ccode = DStoreTranBegin(client, FALSE);
        } else {
            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode == SQLITE_DONE) {
            stmt = client->dstore.sql.sel.dunfo;

            while ((ccode = sqlite3_step(stmt)) == SQLITE_ROW) {
                char iguid[NMAP_GUID_LENGTH+1];

                strncpy(iguid, sqlite3_column_text(stmt, 0), NMAP_GUID_LENGTH);
                iguid[NMAP_GUID_LENGTH] = 0;
                if (NmapCheckAuthorization(client, iguid, NMAP_PRIV_READ, 1)) {
                    continue;
                }

                if ((ccode = ConnWriteF(client->conn, "2001 %s %s %d %d %s\r\n", 
                        sqlite3_column_text(stmt, 0), 
                        sqlite3_column_text(stmt, 1), 
                        sqlite3_column_int(stmt, 2), 
                        sqlite3_column_int(stmt, 3),
                        sqlite3_column_text(stmt, 4))) != -1) {
                    continue;
                }

                break;
            }

            if (ccode == SQLITE_DONE) {
                ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
            } else {
                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }

            sqlite3_reset(stmt);

            DStoreTranEnd(client);
        } else {
            ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
        }

        return(ccode);
    }

    return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
}

int 
NmapCommandDunset(void *param)
{
    int ccode;
    unsigned char *ptr;
    NMAPClient *client = (NMAPClient *)param;
    sqlite3 *handle = client->dstore.handle;
    sqlite3_stmt *stmt;

    /* DUNSET */
    if (client->states & NMAP_CLIENT_USER) {
        if (!(client->states & (NMAP_CLIENT_MBOX | NMAP_CLIENT_CAL))) {
            ptr = client->buffer + 6;
        } else {
            if (client->states & NMAP_CLIENT_MBOX) {
                return(ConnWrite(client->conn, MSG4221MBOXDONE, sizeof(MSG4221MBOXDONE) - 1));
            }

            return(ConnWrite(client->conn, MSG4221CALDONE, sizeof(MSG4221CALDONE) - 1));
        }
    } else {
        return(ConnWrite(client->conn, MSG3241NOUSER, sizeof(MSG3241NOUSER) - 1));
    }

    /* only the user himself can issue this command (for now) */
    if (NMAP_PRINCIPAL_USER != client->principalType || 
        (client->principal[0] && strcmp(client->principal, client->user)))
    {
        return ConnWrite(client->conn, MSG4240NOPERMISSION, 
                         sizeof(MSG4240NOPERMISSION) - 1);
    }

    if (*ptr == '\0') {
        ccode = sqlite3_prepare(handle, NMAP_DUNSET_UPDATE, sizeof(NMAP_DUNSET_UPDATE) - 1, &stmt, NULL);
        if (ccode == SQLITE_OK) {
            ccode = DStoreTranBegin(client, TRUE);
        } else {
            XplConsolePrintf("NMAP [%d]: Prepare update dunset for \"%s\" returned \"%s\"\r\n", __LINE__, client->user, sqlite3_errmsg(handle));

            return(ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1));
        }

        if (ccode == SQLITE_DONE) {
            ccode = sqlite3_step(stmt);

            if (ccode == SQLITE_DONE) {
                ccode = DStoreTranEnd(client);
                if (ccode == SQLITE_DONE) {
                    ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
                } else {
                    ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
                }
            } else {
                DStoreTranAbort(client);

                ccode = ConnWrite(client->conn, MSG5005SQLLIBERR, sizeof(MSG5005SQLLIBERR) - 1);
            }
        } else {
            ccode = ConnWrite(client->conn, MSG4120SQLLOCKED, sizeof(MSG4120SQLLOCKED) - 1);
        }

        sqlite3_finalize(stmt);

        return(ccode);
    }

    return(ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1));
}

static int
DeliverFile(NMAPClient *client, NmapDocumentTypes type, char *sender, char *authSender, char *fullPath)
{
    FILE *fh;
    int ccode;
    struct stat sb;

    if (stat(fullPath, &sb) != 0) {
        return(ConnWrite(client->conn, MSG4201FILENOTFOUND, sizeof(MSG4201FILENOTFOUND) - 1));
    }

    fh = fopen(fullPath, "r");

    if ((ccode = ConnWrite(client->conn, MSG2053SENDRECIPIENTS, sizeof(MSG2053SENDRECIPIENTS) - 1)) == -1) {
        return ccode;
    }
    ConnFlush(client->conn);

    while (ccode != -1) {
        int deliveryStatus;
        char *args[3];
        int numArgs;
        char *recipient;
        char *mbox;
        unsigned long flags;

        ccode = ConnReadLine(client->conn, client->line, CONN_BUFSIZE);
        if ((ccode == -1) || (client->line[0] == '\r' && client->line[1] == '\n')) {
            break;
        }

        numArgs = HulaStringSplit(client->line, ' ', args, 3);
        if (numArgs != 3) {
            ccode = ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1);
            ConnFlush(client->conn);
            continue;
        }

        recipient = args[0];
        mbox = args[1];
        flags = atol(args[2]);

        if (type == NMAP_DOCTYPE_MAIL) {
            deliveryStatus = StoreMessageInMailbox(sender, authSender, recipient, mbox, fh, 0, sb.st_size, flags);
        } else if (type == NMAP_DOCTYPE_MAIL) {
            deliveryStatus = StoreCalendarEvent(sender, authSender, recipient, mbox, fh, sb.st_size, FALSE);
        } else {
            deliveryStatus = DELIVER_UNKNOWN_TYPE;
        }

        fseek(fh, 0, SEEK_SET);

        if (deliveryStatus == DELIVER_SUCCESS) {
            ccode = ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
        } else {
            ccode = ConnWriteF(client->conn, MSG5260DELIVERYFAILED, deliveryStatus);
        }
        ConnFlush(client->conn);
    }

    fclose(fh);

    if (ccode != -1) {
        return ConnWrite(client->conn, MSG1000OK, sizeof(MSG1000OK) - 1);
    } else {
        return ccode;
    }
}
int
NmapCommandDeliverFile(void *param)
{
    NMAPClient *client = (NMAPClient*)param;
    char *ptr;
    char *args[4];
    int numArgs;
    NmapDocumentTypes type;
    char *sender;
    char *authSender;
    char *filename;
    char fullPath[XPL_MAX_PATH + 1];

    if (!(client->states & NMAP_CLIENT_AUTHENTICATED)) {
        return ConnWrite(client->conn, MSG3240NOAUTH, sizeof(MSG3240NOAUTH) - 1);
    }

    ptr = client->buffer + 12;

    /* DELIVER FILE <type> <sender> <authSender> <filename> */
    numArgs = HulaStringSplit(ptr, ' ', args, 4);
    if (numArgs != 4) {
        return ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1);
    }

    type = atoi(args[0]);
    sender = args[1];
    authSender = args[2];
    filename = args[3];

    snprintf(fullPath, sizeof(fullPath), "%s/%s", NMAP.path.spool, filename);

    return DeliverFile(client, type, sender, authSender, fullPath);
}
int
NmapCommandDeliverStream(void *param)
{
    NMAPClient *client = (NMAPClient*)param;
    char *ptr;
    char *args[4];
    int numArgs;
    NmapDocumentTypes type;
    char *sender;
    char *authSender;
    int count;
    int ccode;
    char tmpFile[XPL_MAX_PATH + 1];
    FILE *fh;

    if (!(client->states & NMAP_CLIENT_AUTHENTICATED)) {
        return ConnWrite(client->conn, MSG3240NOAUTH, sizeof(MSG3240NOAUTH) - 1);
    }

    ptr = client->buffer + 14;

    /* DELIVER STREAM <type> <sender> <authSender> <length> */
    numArgs = HulaStringSplit(ptr, ' ', args, 4);
    if (numArgs != 4) {
        return ConnWrite(client->conn, MSG3010BADARGC, sizeof(MSG3010BADARGC) - 1);
    }

    type = atoi(args[0]);
    sender = args[1];
    authSender = args[2];
    count = atol(args[3]);

    fh = XplOpenTemp(MsgGetWorkDir(NULL), "wb", tmpFile);
    if (fh == NULL) {
        return ConnWrite(client->conn, MSG5202TMPWRITEERR, strlen(MSG5202TMPWRITEERR));
    }

    ccode = ConnReadToFile(client->conn, fh, count);
    if (ccode == -1) {
        fclose(fh);
        unlink(tmpFile);
        return ConnWrite(client->conn, MSG5202TMPWRITEERR, strlen(MSG5202TMPWRITEERR));
    }

    if (fclose(fh) != 0) {
        unlink(tmpFile);
        return ConnWrite(client->conn, MSG5202TMPWRITEERR, strlen(MSG5202TMPWRITEERR));
    }

    ccode = DeliverFile(client, type, sender, authSender, tmpFile);

    unlink(tmpFile);

    return ccode;
}
