/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Collections;
using System.Collections.Specialized;
using System.Net;
using System.Net.Sockets;
using System.Web;
using System.IO;
using System.Text;
using Hula.Cal;
using Hula.Store;
using Hula.Util;
using Hula.Nmap;
using Hula.Mdb;
using log4net;
using log4net.Config;

namespace Hula.CalCmd {


public class UnauthorizedPhoneNumberException : System.Exception
{
}


public class CalCmdServer : IHttpHandler  {    
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    static CalCmdServer() {
	Logging.Configure();

	log.Info("CalCmd HTTP Server: started logging");
    }

    public bool IsReusable {
	get { return true; }
    }

    protected string
    FindUser (string phone) 
    {
        if (phone.Length > 3) {
            return "cyrus";
        } else {
            return null;
        }
    }

    
    private void
    QueryCalendar (Command command, IStore store, string user, StringBuilder result)
    {
        StringBuilder url = new StringBuilder(256);
        StringBuilder query = new StringBuilder(256);
        DateTime begin = command.Begin;
        DateTime end = command.End;
        int cnt = 0;

        url.AppendFormat("http://localhost:8081/search/{0}",
                         HttpUtility.UrlEncode(user));

        url.Append("?query=");
        if (CommandType.QueryConflicts == command.Type) {
            DateTime weekago = begin.AddDays(-7);
            DateTime weeklater = begin.AddDays(7);

            query.AppendFormat("hula.store:calendar AND ical.type:VEVENT AND " +
                               "hula.start-time:[{0:D4}{1:D2}{2:D2}{3:D2}{4:D2} TO " +
                               "{5:D4}{6:D2}{7:D2}{8:D2}{9:D2}] AND ",
                               weekago.Year, weekago.Month, weekago.Day,
                               weekago.Hour, weekago.Minute,
                               end.Year, end.Month, end.Day,
                               end.Hour, end.Minute);

            query.AppendFormat("hula.end-time:[{0:D4}{1:D2}{2:D2}{3:D2}{4:D2} TO " +
                               "{5:D4}{6:D2}{7:D2}{8:D2}{9:D2}]",
                               begin.Year, begin.Month, begin.Day,
                               begin.Hour, begin.Minute,
                               weeklater.Year, weeklater.Month, weeklater.Day,
                               weeklater.Hour, weeklater.Minute);
        } else {
            query.AppendFormat("hula.store:calendar AND ical.type:VEVENT AND " +
                               "hula.start-time:[{0:D4}{1:D2}{2:D2}{3:D2}{4:D2} TO " +
                               "{5:D4}{6:D2}{7:D2}{8:D2}{9:D2}]",
                               begin.Year, begin.Month, begin.Day,
                               begin.Hour, begin.Minute,
                               end.Year, end.Month, end.Day,
                               end.Hour, end.Minute);
        }

        url.Append(HttpUtility.UrlEncode(query.ToString()));
        url.Append("&sortkey=hula.start-time&sortorder=ascending&format=cal");
        
        try {
            WebRequest request;
            WebResponse response;
            StreamReader reader;
            string line;
            
            request = WebRequest.Create(url.ToString());
            request.Method = "GET";
            request.Timeout = 10000; /* 10 seconds */
            
            response = request.GetResponse();
            
            reader = new StreamReader(response.GetResponseStream());

            while (null != (line = reader.ReadLine())) {
                string[] parts;
                string summary;
                int tmp;

                if (' ' == line[0]) { /* continuation of previous appt summary */
                    result.Append(line);
                    continue;
                }

                /* line format "(guid)|YYYYMMDDHHMM|YYYYMMDDHHMM|SUMMARY:(text)" */

                parts = line.Split('|');

                begin = DateTime.ParseExact(parts[1], "yyyyMMddHHmm", null);
                end = DateTime.ParseExact(parts[2], "yyyyMMddHHmm", null);

                tmp = parts[3].LastIndexOf(':');
                summary = parts[3].Substring(tmp + 1);
                if (-1 == tmp || null == summary) {
                    throw new FormatException();
                }

                if (0 == cnt && CommandType.QueryConflicts == command.Type) {
                    result.Append("Overlaps with:\r\n");
                }

                Command.PrintConciseDateTimeRange(begin, end, result);
                result.Append(' ');
                result.Append(summary);
                result.Append("\r\n");

                cnt++;
            }
            response.Close();
            
            if (0 == cnt && CommandType.QueryConflicts != command.Type) {
                result.Append("No Events Found.\r\n");
            }

        } catch (ArgumentOutOfRangeException e) {
            result.Append("Server error: " + e.ToString());
        } catch (FormatException e) {
            result.Append("Server error: " + e.ToString());
        } catch (Exception e) {
            result.Append("Server error: " + e.ToString());
        }

    }


    
    public void
    ProcessCalCmd (HttpContext context, string phone, string cmdstr)
    {
        string user;
        Command command;
        StringBuilder message = new StringBuilder(160);
        
        try {

            user = FindUser(phone);
            if (null == user) {
                throw new UnauthorizedPhoneNumberException();
            }

            command = new Command(cmdstr);

            log.DebugFormat ("parsed command: {0}", cmdstr);
            
            log.DebugFormat ("begin: {0}", command.Begin);
            
            log.DebugFormat ("end: {0}", command.End);
            log.DebugFormat ("data: {0}", command.Data);
            log.DebugFormat ("type: {0}", command.Type);
            

            using (IStore store = StoreFactory.GetStore(user, "calendar"))
            {
                string guid;
                string etag;

                switch (command.Type) {
                case CommandType.NewAppt:
                    string icaldata;
                    int len;
                    
                    len = command.ProcessNewAppt(out icaldata);
                    if (0 == len) {
                        throw new HttpException(500, "Internal Error");
                    }

                    store.Add("Personal",
                              new Hashtable(),
                              new StringReader(icaldata),
                              "text/calendar",
                              out guid,
                              out etag);
                    
                    message.AppendFormat("OK, new appt: {0} ", command.Data);
                    command.PrintConciseDateTimeRange(message);
                    message.Append("\r\n");

                    command.Type = CommandType.QueryConflicts;
                    QueryCalendar(command, store, user, message);
                    break;
                
                case CommandType.Query:
                    QueryCalendar(command, store, user, message);
                    break;
                    
                default:
                    throw new HttpException(501, "Not Implemented: " + command.Type); 
                }
            }
        } catch (UnauthorizedPhoneNumberException) {
            message.AppendFormat("Sorry, your phone number is not authorized for " + 
                                 "calendar access.\r\n");
        } catch (InvalidCommandException) {
            message.Append("Sorry, unable to process your message. " +
                           "For help on Hula messages, please " +
                           "send the text 'help'.\r\n");
        }

        context.Response.ContentType = "text/plain";
        context.Response.StatusCode = 200;
        context.Response.Output.Write(message);
    }


    public void 
    ProcessRequest (HttpContext context) 
    {
        string phone = null;
        string command = null;
        NameValueCollection rparams;

	log.DebugFormat("{0} {1} {2}", 
			context.Request.HttpMethod,
			context.Request.Path,
                        context.Request.QueryString);

	try {
	    switch (context.Request.HttpMethod) {
	    case "GET":
                rparams = context.Request.QueryString;
                break;
            case "POST":
                rparams = context.Request.Form;
                break;
	    default :
		throw new HttpException (501, "Not Implemented: " + 
                                         context.Request.HttpMethod);
	    }
            
            phone = rparams["phone"];
            command = rparams["cmd"];
            if (null == phone) {
                throw new HttpException (400, 
                                         "Bad Request: \"phone\" param required.");
            } else if (null == command) {
                throw new HttpException (400, "Bad Request: \"cmd\" param required.");
            }

            ProcessCalCmd (context, phone, command);

	} catch (HttpException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   e.GetHttpCode());
	    log.Warn(e);
	    throw e;
	} catch (NotImplementedException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   501);
	    log.Warn(e);
	    throw new HttpException(501, "Not Implemented", e);
	} catch (Exception e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   500);
	    log.Warn(e);
	    throw e;
	}

	log.InfoFormat("{0} {1}: {2}", 
		       context.Request.HttpMethod,
		       context.Request.Path,
		       context.Response.StatusCode);
    }
}
}
