using System;
using System.Xml;
using System.Web;
using System.Text;
using System.Collections;
using System.IO;
using Lucene.Net.Search;
using Lucene.Net.Index;
using Lucene.Net.Documents;
using Hula.Store;
using Hula.Util;
using Hula.Indexing;
using Hula.Cal;
using log4net;

namespace Hula.Dav {
public abstract class DavReport {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    protected LuceneIndex idx;
    protected IStore store;
    protected string path;
    protected int depth;
    protected string appUrl;
    
    private PropertyFormatter properties = new PropertyFormatter();
    private ArrayList formatters = new ArrayList();
    private TextWriter writer;
    private HttpContext context;

    public static DavReport Get(string type,
				HttpContext context) {
	switch (type) {
	case "calendar-query" :
	    return new CalQueryReport(context);
	case "adbk-query" :
	    return new AddressbookQueryReport(context);
	case "propfind" :
	    return new PropFindReport(context);
 	case "calendar-multiget" :
	case "adbk-multiget" :
	    return new DavMultigetReport(context);
	case "lucene-query" :
	    return new LuceneReport(context);
	default :
	    return null;
	}
    }

    protected DavReport(HttpContext context) {
	this.appUrl = GetAppUrl(context);
	this.writer = context.Response.Output;
	this.context = context;
	AddFormatter(properties);
    }

    private string GetAppUrl(HttpContext context) {
	
	// Let scripts override the app url, in case they're hiding us behind an apache
	string envUrl = Environment.GetEnvironmentVariable("HULAXSP_REPORT_URL");
	if (envUrl != null) {
	    return envUrl;
	}
	
	string protocol = context.Request.ServerVariables["SERVER_PORT_SECURE"];
	if (protocol == null || protocol == "0") {
	    protocol = "http://";
	} else {
	    protocol = "https://";
	}

	string port = context.Request.ServerVariables["SERVER_PORT"];
	if (port == null) {
	    port = "";
	} else {
	    port = ":" + port;
	}

	string host = context.Request.ServerVariables["SERVER_NAME"];	

	return protocol + host + port + context.Request.ApplicationPath;
    }

    public virtual void Setup(LuceneIndex idx, 
			      IStore store,
			      string path,
			      int depth,
			      XmlNode node) {
	this.idx = idx;
	this.store = store;
	this.path = path;
	this.depth = depth;
	Parse(node);
    }

    protected virtual void HandleNode (XmlNode node) {
	switch (node.LocalName) {
	case "prop" :
	    properties.ParseFilter(node);
	    break;
	case "allprop" :
	    properties.AllProps = true;
	    break;
	case "calendar-data" :
	    AddFormatter(new CalendarDataFormatter("C:calendar-data", node));
	    break;
	case "xcalendar-data" :
	    CalFilterXml xmlFilter = new CalFilterXml(node);
	    xmlFilter.IncludeHeader = false;
	    AddFormatter(new StreamFormatter("H:xcalendar-data", xmlFilter));
	    break;
	case "adbk-data" :
	    AddFormatter(new AddressbookDataFormatter("carddav:adbk-data", node));
	    break;
	case "stream" :
	    AddFormatter(new StreamFormatter("H:stream"));
	    break;
	}
    }

    private void Parse(XmlNode node) {
	foreach (XmlNode child in node) {
	    HandleNode(child);
	}
    }

    public abstract void Run();

    public void AddFormatter(ResultFormatter f) {
	formatters.Add(f);
    }

    public void BeginResponse () {
	context.Response.StatusCode = 207;
        context.Response.ContentType = "text/xml; charset=utf-8";
        writer.WriteLine("<?xml version=\"1.0\" encoding=\"utf-8\" ?>");
        writer.WriteLine("<D:multistatus xmlns:D=\"DAV:\" xmlns:C=\"urn:ietf:params:xml:ns:caldav\" xmlns:carddav=\"urn:ietf:params:xml:ns:carddav\" xmlns:H=\"http://hula-project.org/Xml_Namespace/Hula\">");
    }

    public void EndResponse () {
	writer.WriteLine("</D:multistatus>");
    }

    public void Response(IStore store, Document doc, string status) {
	writer.WriteLine("<D:response>");
	writer.WriteLine(" <D:href>{0}{1}</D:href>",
			 appUrl,
			 doc.GetField("hula.davhref").StringValue());
	foreach (ResultFormatter f in formatters) {
	    f.Format(writer, store, doc, status);
	}
	writer.WriteLine("  <D:status>HTTP/1.1 {0}</D:status>", status);
	context.Response.Output.WriteLine("</D:response>");
    }

    public void Response(IStore store, Document doc) {
	Response(store,doc, "200 OK");
    }

    public void Response(IStore store, string status) {
	writer.WriteLine("<D:response>");
	writer.WriteLine(" <D:href>{0}{1}</D:href>",
			 appUrl,
			 "/" + store.Owner + "/" + store.Name + "/");
	foreach (ResultFormatter f in formatters) {
	    f.Format(writer, store, status);
	}

	writer.WriteLine("  <D:status>HTTP/1.1 {0}</D:status>", status);
	context.Response.Output.WriteLine("</D:response>");
    }

    public void Response(IStore store) {
	Response(store, "200 OK");
    }

    public void Responses(IStore store, Hits hits) {
	Hashtable seenGuids = new Hashtable();

	for (int i = 0; i < hits.Length(); i++) {
	    Document doc = hits.Doc(i);
	    string guid = doc.GetField("guid").StringValue();
	    if (seenGuids.Contains(guid)) {
		continue;
	    }

	    seenGuids[guid] = true;
	    Response(store, doc);
	}
    }
}
}
