using System;
using System.IO;
using System.Runtime.InteropServices;

namespace Hula.ICal {


public class ICalParameter : IDisposable {
    IntPtr Handle;
    bool freeHandle;
    string icalString = null;
    string name;
    string value;

    [DllImport(ICalComponent.icalLib)]
    private static extern void icalparameter_free(IntPtr param);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalparameter_as_ical_string(IntPtr property);

    internal ICalParameter(IntPtr param, bool freeHandle) {
	Handle = param;
	this.freeHandle = freeHandle;
    }

    public string Name {
	get {
	    GetValues();
	    return name;
	}
    }

    public string Value {
	get {
	    GetValues();
	    return value;
	}
    }

    private void GetValues() {	
	if (name == null) {
	    if (icalString == null) {
		IntPtr str = icalparameter_as_ical_string(Handle);
		icalString = Marshal.PtrToStringAnsi(str);
	    }
	    string[] pieces = icalString.Split(new char[] {'='}, 2);
	    name = String.Intern(pieces[0]);
	    value = pieces[1];
	}

    }

    public override string ToString() {
	if (icalString == null) {
	    IntPtr str = icalparameter_as_ical_string(Handle);
	    icalString = Marshal.PtrToStringAnsi(str);
	}
	
	return icalString;
    }

    public void Dispose() {
	if (freeHandle && Handle != IntPtr.Zero) {
	    icalparameter_free(Handle);
	    Handle = IntPtr.Zero;
	}
	GC.SuppressFinalize(this);
    }

    ~ICalParameter() {
	Dispose();
    }
}

public class ICalProperty : IDisposable {
    IntPtr Handle;
    bool freeHandle;
    string name = null;
    string value = null;
    string icalString = null;

    [DllImport(ICalComponent.icalLib)]
    private static extern void icalproperty_free(IntPtr property);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalproperty_get_value_as_string(IntPtr property);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalproperty_get_property_name(IntPtr property);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalproperty_as_ical_string(IntPtr property);
    [DllImport(ICalComponent.icalLib)]
    private static extern ICalRecurrence icalproperty_get_rrule(IntPtr property);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalproperty_get_first_parameter(IntPtr property, int kind);
    [DllImport(ICalComponent.icalLib)]
    private static extern IntPtr icalproperty_get_next_parameter(IntPtr property, int kind);


    internal ICalProperty(IntPtr property, bool freeHandle) {
	Handle = property;
	this.freeHandle = freeHandle;
    }

    public string Value {
	get { 
	    if (value == null) {
		IntPtr str = icalproperty_get_value_as_string(Handle); 
		value = Marshal.PtrToStringAnsi(str);
	    }
	    return value;
	}
    }

    public string Name {
	get { 
	    if (name == null) {
		IntPtr str = icalproperty_get_property_name(Handle); 
		name = Marshal.PtrToStringAnsi(str);
	    }
	    return name;
	}
    }

    public ICalRecurrence RecurrenceRule {
	get {
	    return icalproperty_get_rrule(Handle);
	}
    }

    public ICalParameter GetFirstParameter() {
	IntPtr h = icalproperty_get_first_parameter(Handle, 0);
	if (h != IntPtr.Zero) {
	    return new ICalParameter(h, false);
	} else {
	    return null;
	}
    }

    public ICalParameter GetNextParameter() {
	IntPtr h = icalproperty_get_next_parameter(Handle, 0);
	if (h != IntPtr.Zero) {
	    return new ICalParameter(h, false);
	} else {
	    return null;
	}
    }

    public override string ToString() {
	if (icalString == null) {
	    IntPtr str = icalproperty_as_ical_string(Handle);
	    icalString = Marshal.PtrToStringAnsi(str);
	}
	
	return icalString;
    }

    public void Dispose() {
	if (freeHandle && Handle != IntPtr.Zero) {
	    icalproperty_free(Handle);
	    Handle = IntPtr.Zero;
	}
	GC.SuppressFinalize(this);
    }

    ~ICalProperty() {
	Dispose();
    }

    public enum Kind {
	Any = 0,
	Action,
	Allowconflict,
	Attach,
	Attendee,
	CalId,
	CalMaster,
	CalScale,
	CapVersion,
	CarLevel,
	CarId,
	Categories,
	Class,
	Cmd,
	Comment,
	Completed,
	Components,
	Contact,
	Created,
	CsId,
	DateMax,
	DateMin,
	Decreed,
	DefaultCharset,
	DefaultLocale,
	DefaultTzid,
	DefaultVcars,
	Deny,
	Description,
	DtEnd,
	DtStamp,
	DtStart,
	Due,
	Duration,
	ExDate,
	Expand,
	ExRule,
	Freebusy,
	Geo,
	Grant,
	ItipVersion,
	LastModified,
	Location,
	MaxComponentSize,
	MaxDate,
	MaxResults,
	MaxResultsSize,
	Method,
	MinDate,
	Multipart,
	Name,
	Organizer,
	Owner,
	PercentComplete,
	Permission,
	Priority,
	ProdId,
	Query,
	QueryLevel,
	QueryId,
	QueryName,
	Rdate,
	RecurAccepted,
	RecurExpand,
	RecurLimit,
	RecurrenceId,
	RelatedTo,
	RelCalId,
	Repeat,
	Requeststatus,
	Resources,
	Restriction,
	RRule,
	Scope,
	Sequence,
	Status,
	StoresExpanded,
	Summary,
	Target,
	Transp,
	Trigger,
	Tzid,
	Tzname,
	Tzoffsetfrom,
	Tzoffsetto,
	Tzurl,
	Uid,
	Url,
	Version,
	X,
	Xlicclass,
	Xlicclustercount,
	Xlicerror,
	Xlicmimecharset,
	Xlicmimecid,
	Xlicmimecontenttype,
	Xlicmimeencoding,
	Xlicmimefilename,
	Xlicmimeoptinfo,
	No
    }
}
}
