/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Collections;
using System.Web;
using System.IO;

using Hula.Nmap;
using Hula.Util;

using Hula.Importing;
using Hula.Store;

using log4net;

namespace Hula.Import {

public class WrongUserException : System.Exception {
}

public class ImportService : IHttpHandler {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    static ImportService () {
	Logging.Configure ();
	log.Info ("started logging");
    }

    public bool IsReusable {
	get { return true; }
    }

    public static void DecomposeUrl(string url,
				    out string owner,
				    out string store,
				    out string path)
    {
	owner = store = path = null;

	url = url.Substring(1);

	int i;
	i = url.IndexOf('/');

	if (i > -1) {
	    owner = url.Substring(0, i);
	    url = url.Substring(i + 1);
	} else {
	    owner = url;
	    store = null;
	    path = null;
	    return;
	}

	i = url.IndexOf('/');

	if (i > 0) {
	    store = url.Substring(0, i);
	    path = url.Substring(i);
	} else {
	    store = url;
	    path = null;
	}
    }    

    private string GetLocalPath(HttpRequest request, string path) {
	if (path.StartsWith (request.ApplicationPath)) {
	    return path.Substring(request.ApplicationPath.Length).TrimEnd(new char[] { '/' });
	} else {
	    return path.TrimEnd(new char[] { '/' });
	}
    }

    private void SplitPath(string path, out string folder, out string filename) {
        int i = path.LastIndexOf('/');
        if (i > 0) {
            folder = path.Substring(0, i);
            filename = path.Substring(i + 1);
        } else {
            folder = "/";
            filename = path.Substring(1);
        }
    }

    private void HandleOptions(HttpContext context, string user, string path) {
	context.Response.StatusCode = 200;
	context.Response.AddHeader("Allow", "OPTIONS, PUT, POST");
    }

    private void HandlePut(HttpContext context, string user, string href) {
	string owner, storeName, path;
	string folder, filename;

	DecomposeUrl (href, out owner, out storeName, out path);

	if (user != owner) {
	    throw new WrongUserException ();
	}

	SplitPath (path, out folder, out filename);

	using (IStore store = StoreFactory.GetStore (owner, storeName)) {
	    Importer importer = Importer.GetImporter (filename,
						      context.Request.ContentType.ToLower (),
						      context.Request.InputStream);

	    if (importer == null) {
		throw new HttpException (405, "Method not allowed");
	    }

	    IList guids = importer.ImportTo (store, folder);

	    context.Response.StatusCode = 200;
	    context.Response.ContentType = "text/plain; charset=utf-8";

	    foreach (string guid in guids) {
		context.Response.Output.WriteLine (guid);
	    }
	}
    }

    private void HandlePost(HttpContext context, string user, string href) {
	string owner, storeName, path;
	string folder, filename;
        bool successes = false;
        bool failures = false;

	DecomposeUrl (href, out owner, out storeName, out path);
        if (user != owner) {
            log.InfoFormat ("user {0} != owner {1}", user, owner);
            throw new WrongUserException ();
        }

        if (!context.Request.ContentType.StartsWith("multipart/form-data")) {
            throw new HttpException (405, "Method not allowed");
        }

        if (context.Request.Files == null) {
            /* no work to do */
            return;
        }

	SplitPath (path, out folder, out filename);

        log.InfoFormat ("got folder: {0}", folder);
        context.Response.ContentType = "text/plain";

        using (IStore store = StoreFactory.GetStore (owner, storeName)) {
            string[] files = context.Request.Files.AllKeys;
            for (int i = 0; i < files.Length; i++) {
                HttpPostedFile file = context.Request.Files[files[i]];
                Importer importer = Importer.GetImporter (file.FileName,
                                                          file.ContentType.ToLower (),
                                                          file.InputStream);
                
                if (importer == null) {
                    log.WarnFormat ("Could not import {0} file {1}",
                                    file.ContentType, file.FileName);
                    failures = true;
                    continue;
                }

                context.Response.Output.WriteLine ("{0}:", file.FileName);

                foreach (string guid in importer.ImportTo (store, folder)) {
                    successes = true;
                    // log.InfoFormat ("{0} -> {1}", file.FileName, guid);
                    context.Response.Output.WriteLine ("    {0}", guid);
                }
            }
        }

        context.Response.StatusCode = successes ? 200 : 400;
    }

    public void ProcessRequest (HttpContext context) {
        string path = GetLocalPath(context.Request, context.Request.Path);
	string user = context.User.Identity.Name;

	log.DebugFormat("{0} {1}", 
			context.Request.HttpMethod,
			context.Request.Path);

	try {
	    switch (context.Request.HttpMethod) {
	    case "OPTIONS" :
		HandleOptions(context, user, path);
		break;
	    case "PUT" :
		HandlePut(context, user, path);
		break;
	    case "POST":
		HandlePost(context, user, path);
		break;
	    default :
		throw new HttpException (501, "Not Implemented: " + context.Request.HttpMethod);
	    }
	} catch (WrongUserException e) {
	    context.Response.AppendHeader ("WWW-Authenticate", "Basic realm=\"Hula\"");
	    context.Response.StatusCode = 401;
	    context.Response.StatusDescription = "Access Denied";
	    context.Response.Write ("401 Access Denied");

	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   context.Response.StatusCode);
	    log.Warn(e);
	} catch (NmapException e) {
	    Exception newE = Hula.Nmap.NmapClient.NmapToHttpException(e);
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   400);
	    log.Warn(newE);
	    throw newE;
	} catch (HttpException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   e.GetHttpCode());
	    log.Warn(e);
	    throw e;
	} catch (NotImplementedException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   501);
	    log.Warn(e);
	    throw new HttpException(501, "Not Implemented", e);
	} catch (Exception e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   context.Request.HttpMethod,
			   context.Request.Path,
			   500);
	    log.Warn(e);
	    throw e;
	}

	log.InfoFormat("{0} {1}: {2}", 
		       context.Request.HttpMethod,
		       context.Request.Path,
		       context.Response.StatusCode);
    }
}

}
