using System;
using System.Collections;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Threading;

using log4net;

using Hula.Mdb;
using Hula.Util;
using Lucene.Net.Documents;
using Lucene.Net.Search;

namespace Hula.Indexer
{
public class IndexServer : IDisposable
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
    private Mdb.Mdb mdb;
    private IList trustedHosts;

    public IndexServer() {
        this.mdb = new Mdb.Mdb(null, null);

        log.Debug("getting trusted host list");
        this.trustedHosts = GetTrustedHosts();
    }

    public void Dispose() {
        mdb.Dispose();
    }

    private class Handler  {
        Socket socket;
        IndexClient indexClient;
        IList trustedHosts;

        private static Char[] argDelim = " ".ToCharArray();

        public Handler(IndexClient indexClient, Socket socket, IList trustedHosts) {
            this.indexClient = indexClient;
            this.socket = socket;
            this.trustedHosts = trustedHosts;
        }
        
        public void Run() {
            using (socket) {
                try {
                    NetworkStream s = new NetworkStream(socket);
                    StreamReader reader = new StreamReader(s);
                    StreamWriter writer = new StreamWriter(s);

                    if (!Authenticate(reader, writer)) {
                        log.Info("Authentication failed");
                        return;
                    }

                    string line = reader.ReadLine();

                    while (line != null) {
                        log.DebugFormat("Handling: {0}", line);
                        string[] cmd = line.Split(argDelim, 2);

                        if (cmd.Length < 2) {
                            writer.WriteLine("No command or args!");
                            writer.Flush();
                        }

                        switch (cmd[0]) {
                        case "INDEX":
                            Index(cmd[1], reader, writer);
                            break;
                        case "SEARCH":
                            Search(cmd[1], reader, writer);
                            break;
                        default:
                            break;
                        }
                        line = reader.ReadLine();
                    }
                } catch (Exception e) {
                    log.ErrorFormat("Exception handling request: {0}", e);
                }
            }
        }

        public bool Authenticate(StreamReader reader, StreamWriter writer) {
            IPEndPoint remote = (IPEndPoint)socket.RemoteEndPoint;
            string host;
            try {
                host = Dns.GetHostByAddress(remote.Address).HostName;
            } catch (SocketException e) {
                host = remote.Address.ToString ();
            }
            if (trustedHosts.Contains(remote.Address.ToString())) {
                log.DebugFormat("host {0} is trusted", remote.Address);
                writer.WriteLine("1000 {0} Hula NMAP Indexing server ready",
                                 host);
                writer.Flush();
                return true;
            }

            int thread = Thread.CurrentThread.GetHashCode() & 0xFFFF;
            int time = Convert.ToInt32((DateTime.Now - new DateTime (1970, 1, 1)).TotalSeconds);

            string salt = String.Format("<{0:X}{1}{2:X}>", Convert.ToString(thread), host, time);

            writer.WriteLine("4242 NMAP {0}", salt);
            writer.Flush();

            string line = reader.ReadLine();

            if (line.StartsWith("AUTH ")) {
                string creds = line.Substring(5);
                string hash = MsgApi.NmapChallenge(salt);

                if (hash.Equals(creds)) {
                    writer.WriteLine("1000 OK");
                    writer.Flush();
                    return true;
                }
            }

            log.Debug("Failure.");
            return false;
        }

        public void Index(string cmd, StreamReader reader, StreamWriter writer) {
            string[] args = cmd.Split(argDelim);

            if (args.Length < 3) {
                writer.WriteLine("ERROR Too few arguments");
                writer.Flush();
                return;
            }

            string user = args[0];
            string store = args[1];
            HulaThreadPriority priority;
            string[] guids = null;

            try
            {
                priority = (HulaThreadPriority)HulaThreadPriority.Parse(typeof(HulaThreadPriority), args[2]);
            }
            catch (ArgumentException e)
            {
                writer.WriteLine("ERROR Invalid priority");
                writer.Flush();
                return;
            }

            int guidStart = 3;

            if (args.Length > guidStart)
            {
                guids = new string[args.Length - guidStart];
                for (int i=guidStart; i<args.Length; i++)
                {
                    guids[i-guidStart] = args[i];
                }
            }

            if (priority == HulaThreadPriority.Interactive)
            {
                log.InfoFormat("Running blocking index request: {0}, {1}", user, store);
                indexClient.IndexUser(user, store, guids, priority);
            } else {
                log.InfoFormat("Queueing index request: {0}, {1} (priority {2})", user, store, priority);
                indexClient.QueueIndexUser(user, store, guids, priority);
            }

            writer.WriteLine("1000 OK");
            writer.Flush();
        }

        public void Search(string cmd, StreamReader reader, StreamWriter writer) {
            string[] args = cmd.Split(argDelim, 2);

            if (args.Length < 2) {
                writer.WriteLine("ERROR Too few arguments");
                writer.Flush();
                return;
            }

            string user = args[0];
            string query = args[1];

            IList docs = indexClient.SearchUser(user, query);

            if (docs.Count == 0) {
                writer.WriteLine("No results");
            }

            foreach (IDictionary doc in docs) {
                writer.WriteLine("{0} {1}", doc["guid"], doc["summary"]);
            }

            writer.Flush();
        }
    }

    public void Run(IndexClient indexClient) {
        log.Debug("running server");
        IPAddress addr = IPAddress.Parse("127.0.0.1");
        TcpListener server = new TcpListener(addr, 8689);

        try {
            server.Start();
        } catch (SocketException e) {
            log.ErrorFormat("cannot start server: {0}", e.Message);
            return;
        }

        indexClient.IndexNMAPStore();
        
        while (true) {
            Socket socket = server.AcceptSocket();
            log.Debug("client connected");
        
            Handler h = new Handler(indexClient, socket, trustedHosts);
            Thread t = new Thread(new ThreadStart(h.Run));
            t.Start();
        }
    }

    public void HandleClient(object o) {
        
    }

    private IList GetTrustedHosts() {
        string trustDN = "\\Tree\\Internet Services\\Hula Messaging Server\\NMAP Agent";

        if (!mdb.IsInitialized) {
            log.Error("MDB isn't initialized!");
        }

        if (!mdb.IsAuthenticated) {
            log.Error("MDB isn't authenticated!");
        }

        if (!mdb.ObjectExists(trustDN)) {
            log.ErrorFormat("No MDB object found: {0}", trustDN);
            return new ArrayList();
        } else {
            using (MdbObject obj = mdb.GetObject(trustDN)) {
                return new ArrayList(obj.GetAttribute("Novonyx:NMAP Trusted Hosts"));
            }
        }
    }
}
}
