using System;
using System.IO;
using System.Collections;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using Hula.Nmap;
using Hula.Cal;
using log4net;
using log4net.Config;

namespace Hula.Store {

public class NmapStore : FilteringStore, IStore, IDisposable {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    private static Hashtable DocumentTypeMimeTypes = new Hashtable();

    static NmapStore() {
	DocumentTypeMimeTypes[DocumentType.Unknown]     = "application/octet-stream";
	DocumentTypeMimeTypes[DocumentType.Mail]        = "message/rfc822";
	DocumentTypeMimeTypes[DocumentType.Calendar]    = "text/calendar";
	DocumentTypeMimeTypes[DocumentType.Addressbook] = "text/x-vcard";

	DocumentTypeMimeTypes[DocumentType.UnknownFolder]     = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.MailFolder]        = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.CalendarFolder]    = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.AddressbookFolder] = "application/x-directory";

	DocumentTypeMimeTypes[DocumentType.UnknownSharedFolder]     = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.MailSharedFolder]        = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.CalendarSharedFolder]    = "application/x-directory";
	DocumentTypeMimeTypes[DocumentType.AddressbookSharedFolder] = "application/x-directory";
    }

    private string owner;
    private string storeAlias;
    private DocumentType docType;
    private DocumentType folderType;

    private NmapClient nmap;

    public string Owner { get { return owner; } }
    public string Name { get { return storeAlias; } }

    private void SetTypes(string storeName) {
	switch (storeName) {
	case "calendar" :
	    docType = DocumentType.Calendar;
	    folderType = DocumentType.CalendarFolder;
	    break;
	case "mail" :
	    docType = DocumentType.Mail;
	    folderType = DocumentType.MailFolder;
	    break;
	case "addressbook" :
	    docType = DocumentType.Addressbook;
	    folderType = DocumentType.AddressbookFolder;
	    break;
	default :
	    docType = DocumentType.Unknown;
	    folderType = DocumentType.UnknownFolder;
	    break;
	}
    }

    public NmapStore(string owner, string storeName) {
        this.owner = owner;
	this.storeAlias = storeName;

	nmap = new NmapClient(owner);
	nmap.Auth.USER(owner);

	SetTypes(storeName);
    }

    public NmapStore(string owner, string storeName, NmapClient nmap) {
        this.owner = owner;
	this.storeAlias = storeName;

	this.nmap = nmap;

	SetTypes(storeName);	
    }
    
    private string FolderToResource(string folder) {
	folder = folder.Trim('/');
	return nmap.EncodeResource(folder);
    }

    private string ResourceToFolder(string resource) {
	return "/" + nmap.DecodeResource(resource);
    }

    private bool NmapTypeIsFolder(DocumentType type) {
	return ((type & (DocumentType)DocumentTypeMasks.Folder) != 0);
    }

    public void Add (string folder, 
		     IDictionary properties,
		     TextReader reader,
		     string mimeType,
		     out string guid,
		     out string etag) {
	string resource;
	Hula.Util.Filter filter;

	resource = FolderToResource(folder);

	if (resource == "") {
	    throw new System.Web.HttpException(405, "Method not allowed");
	}

	if (!GetFilter (mimeType, (string)DocumentTypeMimeTypes[docType], out filter, out mimeType)) {
	    log.Warn ("Could not convert from " + mimeType + " to " + docType + " (" + (string)DocumentTypeMimeTypes[docType] + ")");
	    throw new System.Web.HttpException(405, "Method not allowed");
	}

	nmap.Document.DSTOR(resource, docType, reader, filter, out guid, out etag);

	// ETags must be surrounded by quotes
	etag = '"' + etag + '"';

	// FIXME: setting initial properties should be done atomically with the DSTOR
	SetProperties(guid, properties);
	SetProperty(guid, "hula.mime-type", mimeType);
    }

    public TextReader Get (string guid) {
	return nmap.Document.DREAD(guid);
    }

    public void Get(string guid, out IDictionary properties, out TextReader reader) {
	NmapStreamReader nmapReader;
	nmap.Document.DLIST(guid, out properties, out nmapReader);
	reader = nmapReader;
	
	NormalizeProperties(properties);
    }

    public void Replace (string guid, string etag, TextReader reader, string mimeType, out string newEtag) {
	Hula.Util.Filter filter;

	if (etag != null) {
	    etag = etag.Trim('"');
	}

	if (!GetFilter (mimeType, (string)DocumentTypeMimeTypes[docType], out filter, out mimeType)) {
	    log.Warn ("Could not convert from " + mimeType + " to " + docType);
	    throw new System.Web.HttpException (405, "Method not allowed");
	}
	
	try {
	    nmap.Document.DREPL(guid, etag, reader, filter, out newEtag);
	    SetProperty(guid, "hula.mime-type", mimeType);
	    // ETags must be surrounded by quotes
	    newEtag = '"' + newEtag + '"';

	} catch (NmapException e) {
	    if (e.Code == 4226 || e.Code == 4221) {
		throw new ETagException(e);
	    } else {
		throw e;
	    }
	}
    }    

    public void Move (string guid, string etag, string newFolder) {
	throw new NotImplementedException();
    }
    
    public string Copy (string guid, string etag, string newFolder) {
	throw new NotImplementedException();
    }
    
    public void Delete (string guid, string etag) {
	try {
	    nmap.Document.DPURG(guid);
	} catch (NmapException e) {
	    if (e.Code == 4226 || e.Code == 4221) {
		throw new ETagException(e);
	    } else {
		throw e;
	    }
	}
    }
    
    public void NormalizeProperties(IDictionary properties) {
	DocumentType documentType = (DocumentType)Int32.Parse((string)properties["nmap.type"]);

	// Normalize some properties
	properties["guid"] = properties["nmap.guid"];
	properties.Remove("nmap.guid");

	if (properties.Contains("nmap.length")) {
	    properties["hula.length"] = properties["nmap.length"];
	    properties.Remove("nmap.length");
	}

	// Calculate properties that are based on nmap proeprties
	if (!properties.Contains("hula.mime-type")) {
	    properties["hula.mime-type"] = (string)DocumentTypeMimeTypes[documentType];
	}
	
	properties["hula.etag"] = '"' + (string)properties["nmap.modified"] + '"';
	
	string folderName = ResourceToFolder((string)properties["nmap.resource"]);
	if (NmapTypeIsFolder (documentType)) {
	    string parent, filename;
	    SplitPath(folderName, out parent, out filename);
	    properties["hula.folder"] = parent;
	    properties["hula.davfilename"] = filename;
	} else {
	    properties["hula.folder"] = folderName;
	}

	if (!properties.Contains("hula.store")) {
	    properties["hula.store"] = storeAlias;
	}
    }

    public IDictionary GetProperties (string guid) {
	IDictionary properties = nmap.Document.DPGET(guid);

	NormalizeProperties(properties);
	return properties;
    }

    public string GetProperty (string guid, string key) {
	IDictionary dict = GetProperties(guid);
	return (string)dict[key];
    }

    private void SetProperties (string guid, IDictionary dict) {
	foreach (string key in dict.Keys) {
	    SetProperty(guid, key, (string)dict[key]);
	}
    }
    
    public void SetProperty (string guid, string key, string value) {
	nmap.Document.DPSET(guid, key, value);
    }

    private void SplitPath(string path, out string folder, out string filename) {
        int i = path.LastIndexOf('/', 1);
        if (i > 0) {
            folder = path.Substring(0, i);
            filename = path.Substring(i + 1);
        } else {
            folder = "/";
            filename = path.Substring(1);
        }
    }

    public string AddFolder(string folder) {
        string resource = FolderToResource(folder);
	string guid, etag;

	try {
	    nmap.Document.DCREA(folderType, resource, out guid, out etag);
	} catch (NmapException e) {
	    if (e.Code == 4226 || e.Code == 4221) {
		throw new ETagException(e);
	    } else {
		throw e;
	    }
	}
	
        return guid;
    }

    public void Dispose() {
	nmap.Dispose();
	GC.SuppressFinalize(this);
    }

    ~NmapStore() {
	nmap.Dispose();
    }
}
}
