/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2001-2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" and
 *    "Apache Lucene" must not be used to endorse or promote products
 *    derived from this software without prior written permission. For
 *    written permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    "Apache Lucene", nor may "Apache" appear in their name, without
 *    prior written permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

using Lucene.Net.Index;
using Lucene.Net.Search;
using System.Collections;

namespace Hula.Util {

public class ChainedFilter : Lucene.Net.Search.Filter {
    public enum Operation {
	NoOp,
	Or,
	And,
	Xor
    };

    private ArrayList chain = null;
    private ArrayList logicArray = null;
    private Operation logic = Operation.NoOp;
    
    public ChainedFilter(ArrayList chain, ArrayList logicArray) {
	this.chain = chain;
	this.logicArray = logicArray;
    }

    public ChainedFilter(ArrayList chain, Operation logic) {
	this.chain = chain;
	this.logic = logic;
    }

    public override System.Collections.BitArray Bits(IndexReader reader) {
	if (logic != Operation.NoOp) {
	    return Bits(reader, logic);
	} else if (logicArray != null) {
	    return Bits(reader, logicArray);
	} else {
	    return Bits(reader, Operation.And);
	}
    }

    private System.Collections.BitArray Bits(IndexReader reader, Operation logic) {
	BitArray result;
	int i = 0;
	
	if (logic == Operation.And) {
	    result = (BitArray)((Lucene.Net.Search.Filter)chain[i]).Bits(reader).Clone();
	    i++;
	} else {
	    result = new System.Collections.BitArray((reader.MaxDoc() % 64 == 0?reader.MaxDoc() / 64:reader.MaxDoc() / 64 + 1) * 64);
	}

	foreach (Lucene.Net.Search.Filter f in chain) {
	    DoChain(result, reader, logic, f);
	}

	return result;
    }

    private System.Collections.BitArray Bits(IndexReader reader, ArrayList logicChain) {
        if (chain.Count != logicChain.Count) {
            throw new System.ArgumentException("Invalid number of elements in logic array");
	}

	BitArray result;

	int i = 0;
	
	if ((Operation)logicChain[i] == Operation.And) {
	    result = (BitArray)((Lucene.Net.Search.Filter)chain[i]).Bits(reader).Clone();
	    i++;
	} else {
	    result = new System.Collections.BitArray((reader.MaxDoc() % 64 == 0?reader.MaxDoc() / 64:reader.MaxDoc() / 64 + 1) * 64);
	}

	for (; i < chain.Count; i++) {
	    DoChain(result, reader, (Operation)logicChain[i], (Lucene.Net.Search.Filter)chain[i]);
	}

	return result;
    }

    private void DoChain(BitArray result, IndexReader reader,
                         Operation logic, Lucene.Net.Search.Filter filter) {
        switch (logic) {
            case Operation.Or:
                result.Or(filter.Bits(reader));
                break;
            case Operation.And:
                result.And(filter.Bits(reader));
                break;
            case Operation.Xor:
                result.Xor(filter.Bits(reader));
                break;
            default:
                DoChain(result, reader, Operation.Or, filter);
                break;
        }
    }
}    
}
