/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

/** \file nmap.c
 */

#include <config.h>
#include <xpl.h>
#include <hulautil.h>
#include <openssl/md5.h>

#include <memmgr.h>

#include <nmlib.h>
#include <msgapi.h>
#include <nmap.h>

struct {
    RegistrationStates state;

    unsigned char access[NMAP_HASH_SIZE];

    MDBHandle directoryHandle;

    SSL_CTX *context;

    BOOL debug;
} NMAPLibrary = {
    REGISTRATION_LOADING, 
    { '\0' }, 
    NULL, 
    NULL, 

    FALSE
};

int 
NMAPSendCommand(Connection *conn, const unsigned char *command, size_t length)
{
    register const char *b;
    register size_t i;
    register size_t r;

    b = command;
    r = length;
    if (r > 0) {
        if (r < conn->send.remaining) {
            memcpy(conn->send.write, b, r);

            conn->send.write += r;
            conn->send.remaining -= r;

            CONN_TCP_FLUSH(conn, conn->send.read, conn->send.write, i);
            if (i > 0) {
                conn->send.read = conn->send.write = conn->send.buffer;
                conn->send.remaining = CONN_TCP_MTU;

                return(length);
            }
        } else {
            if (conn->send.remaining) {
                memcpy(conn->send.write, b, conn->send.remaining);
                b += conn->send.remaining;
                r -= conn->send.remaining;

                conn->send.write += conn->send.remaining;
                conn->send.remaining = 0;
            }

            CONN_TCP_FLUSH(conn, conn->send.read, conn->send.write, i);
            if (i > 0) {
                conn->send.read = conn->send.write = conn->send.buffer;
                conn->send.remaining = CONN_TCP_MTU;

                CONN_TCP_FLUSH(conn, b, b + r, i);
                if (i > 0) {
                    return(length);
                }
            }
        }
    }

    return((length == 0)? 0: -1);
}

int
NMAPSendCommandF(Connection *conn, const char *format, ...)
{
    int i;
    va_list ap;
    Connection *c = conn;

    do {
        if (CONN_BUFSIZE < c->send.remaining) {
            va_start(ap, format);
            i = XplVsnprintf(c->send.write, CONN_BUFSIZE, format, ap);
            va_end(ap);

            if (i >= 0) {
                c->send.write += i;
                c->send.remaining -= i;

                CONN_TCP_FLUSH(c, c->send.read, c->send.write, i);
                if (i > 0) {
                    c->send.read = c->send.write = c->send.buffer;
                    c->send.remaining = CONN_TCP_MTU;

                    return(i);
                }
            }
        }

        CONN_TCP_FLUSH(c, c->send.read, c->send.write, i);
        if (i > 0) {
            c->send.read = c->send.write = c->send.buffer;
            c->send.remaining = CONN_TCP_MTU;

            continue;
        }

        return(-1);
    } while (TRUE);
}

__inline static unsigned char *
FindNewLineChar(unsigned char *Buffer, unsigned char *EndPtr) {
    register unsigned char *ptr = Buffer;
    register unsigned char *endPtr = EndPtr;

    do {
        if (ptr < endPtr) {
            if (*ptr != '\n') {
                ptr++;
                continue;
            }

            return(ptr);
        }

        break;
    } while(TRUE);

    return(NULL);
}

__inline static unsigned char *
FindEndOfLine(Connection *conn)
{
    int count;
    char *newLine;

    Connection *c = conn;

    do {
        if ((newLine = FindNewLineChar(c->receive.read, c->receive.write)) != NULL) {
            return(newLine);
        }

        count = c->receive.write - c->receive.read;

        if (count < CONN_TCP_MTU) {
            if (count == 0) {
                CONN_TCP_RECEIVE(c, c->receive.buffer, CONN_TCP_MTU, count);
                if (count > 0) {
                    c->receive.read = c->receive.buffer;
                    c->receive.write = c->receive.buffer + count;
                    c->receive.remaining = CONN_TCP_MTU - count;
                    c->receive.write[0] = '\0';
                    continue;
                }

                break;
            }

            /* preserve unread data in the buffer */
            memcpy(c->receive.buffer, c->receive.read, count);
            c->receive.read = c->receive.buffer;
            c->receive.write = c->receive.buffer + count;
            c->receive.remaining = CONN_TCP_MTU - count;
            CONN_TCP_RECEIVE(c, c->receive.write, c->receive.remaining, count);
            if (count > 0) {
                c->receive.write += count;
                c->receive.remaining -= count;
                c->receive.write[0] = '\0';
                continue;
            }

            break;
        }

        /* This line is longer than MTU size, NMAP is not supposed to do that on a response line */
        return(c->receive.write - 1);
    } while (TRUE);

    return(NULL);
}

__inline static int 
NMAPReadDataAfterResponseCode(Connection *conn, unsigned char *response, size_t length, BOOL RemoveCRLF)
{
    int ccode;
    int count;
    char *newLine;

    do {
        newLine = FindEndOfLine(conn);
        if (newLine) {
            if (*newLine == '\n') {
                newLine++;
                ccode = atol(conn->receive.read);
                if (ccode != 6000) {
                    if ((ccode > 999) && (ccode < 10000) && ((conn->receive.read[4] == ' ') || (conn->receive.read[4] == '-'))) {
                        if (RemoveCRLF) {
                            if (newLine[-2] == '\r') {
                                count = newLine - conn->receive.read - 7;
                            } else {
                                count = newLine - conn->receive.read - 6;
                            }
                        } else {
                            count = newLine - conn->receive.read - 5;
                        }

                        if ((unsigned long)count < length) {
                            memcpy(response, conn->receive.read + 5, count);
                            response[count] = '\0';
                            conn->receive.read = newLine;
                            return(ccode);
                        }

                        return(NMAP_ERROR_BUFFER_TOO_SHORT);
                    } 

                    return(NMAP_ERROR_INVALID_RESPONSE_CODE);
                }

                if (conn->client.cb) {
                    if (((NMAPOutOfBoundsCallback)conn->client.cb)(conn->client.data, conn->receive.read + 5, newLine)) {
                        conn->receive.read = newLine;
                        continue;
                    }

                    return(NMAP_ERROR_CALLBACK_FUNCTION_FAILED);
                }

                return(NMAP_ERROR_CALLBACK_FUNCTION_NOT_DEFINED);
            }

            return(NMAP_ERROR_LINE_TOO_LONG);
        }

        break;
    } while (TRUE);

    return(NMAP_ERROR_COMM);
}

__inline static int 
NMAPReadFullLine(Connection *conn, unsigned char *response, size_t length, BOOL RemoveCRLF)
{
    int ccode;
    int count;
    char *newLine;

    do {
        newLine = FindEndOfLine(conn);
        if (newLine) {
            if (*newLine == '\n') {
                newLine++;
                ccode = atol(conn->receive.read);
                if (ccode > -1) {
                    if (ccode != 6000) {
                        if (RemoveCRLF) {
                            if (newLine[-2] == '\r') {
                                count = newLine - conn->receive.read - 2;
                            } else {
                                count = newLine - conn->receive.read - 1;
                            }
                        } else {
                            count = newLine - conn->receive.read;
                        }

                        if ((unsigned long)count < length) {
                            memcpy(response, conn->receive.read, count);
                            response[count] = '\0';
                            conn->receive.read = newLine;
                            return(ccode);
                        }

                        return(NMAP_ERROR_BUFFER_TOO_SHORT);
                    }

                    if (conn->client.cb) {
                        if (((NMAPOutOfBoundsCallback)conn->client.cb)(conn->client.data, conn->receive.read + 5, newLine)) {
                            conn->receive.read = newLine;
                            continue;
                        }

                        return(NMAP_ERROR_CALLBACK_FUNCTION_FAILED);
                    }

                    return(NMAP_ERROR_CALLBACK_FUNCTION_NOT_DEFINED);
                }

                return(NMAP_ERROR_INVALID_RESPONSE_CODE);
            }

            return(NMAP_ERROR_LINE_TOO_LONG);
        }

        break;
    } while (TRUE);

    return(NMAP_ERROR_COMM);
}


__inline static int 
NMAPReadJustResponseCode(Connection *conn)
{
    int ccode;
    char *newLine;

    do {
        newLine = FindEndOfLine(conn);
        if (newLine) {
            if (*newLine == '\n') {
                newLine++;
                ccode = atol(conn->receive.read);
                if (ccode != 6000) {
                    if ((ccode > 999) && (ccode < 10000) && ((conn->receive.read[4] == ' ') || (conn->receive.read[4] == '-'))) {
                        conn->receive.read = newLine;
                        return(ccode);
                    } 

                    return(NMAP_ERROR_INVALID_RESPONSE_CODE);
                }

                if (conn->client.cb) {
                    if (((NMAPOutOfBoundsCallback)conn->client.cb)(conn->client.data, conn->receive.read + 5, newLine)) {
                        conn->receive.read = newLine;
                        continue;
                    }

                    return(NMAP_ERROR_CALLBACK_FUNCTION_FAILED);
                }

                return(NMAP_ERROR_CALLBACK_FUNCTION_NOT_DEFINED);
            }

            return(NMAP_ERROR_LINE_TOO_LONG);
        }

        break;
    } while (TRUE);

    return(NMAP_ERROR_COMM);
}

/** Used to read response lines from NMAP.

    Remarks
        - will block until a line can be read from the network
        - for every 6000 line encountered the callback function is called and the next line is found
        - copies the first non 6000 line excluding the \\r\\n into the response buffer
        - advances read pointer to the beginning of the next line on success

    @return 
        - positive numeric equivalent of the first characters on the line on success
        - negative number on failure

    @param conn the network connection to read the line from

    @param response is either a pointer to a buffer or NULL.
        - \<destination buffer\>
            - the response code will be returned
            - the line will be copied to the destination buffer
        - NULL
            - the response code will be returned

    @param length the length of the destination buffer

    @param check is a boolean value with the following effects:
        - TRUE 
            - the NMAP response code will be validated
            - only the data after the response code will be copied to the destination buffer
        - FALSE 
            - will not require the first characters read to be an NMAP response code
            - the entire line will be copied to the destination buffer

 */
int 
NMAPReadResponse(Connection *conn, unsigned char *response, size_t length, BOOL check)
{
    if (response) {
        if (check) {
            return(NMAPReadDataAfterResponseCode(conn, response, length, TRUE));
        } else {
            return(NMAPReadFullLine(conn, response, length, TRUE));
        }
    } else {
        return(NMAPReadJustResponseCode(conn));
    }
}


/* NMAPReadResponseLine

   Same as NMAPReadResponse except that it does not strip the \r\n 

 */
int 
NMAPReadResponseLine(Connection *conn, unsigned char *response, size_t length, BOOL check)
{
    if (response) {
        if (check) {
            return(NMAPReadDataAfterResponseCode(conn, response, length, FALSE));
        } else {
            return(NMAPReadFullLine(conn, response, length, FALSE));
        }
    } else {
        return(NMAPReadJustResponseCode(conn));
    }
}

/* NMAPReadResponseAndCount

    Remarks
        This function is designed for reading NMAP responses that take the form:
        
        <response code><' ' | '-'><count>
        
        2021 374                     (for example)

        This function:
        - will block until a line can be read from the network
        - for every 6000 line encountered the callback function is called and the next line is found
        - evaluates the first non 6000 line 
        - advances read pointer to the beginning of the next line on success

    Return Value
        - the NMAP response code on success
        - negative number on failure

    Parameters
        conn    - the network connection to read the line from

        count   - this variable will be assigned the numberic equivalent of 
                  the string following the NMAP response code.
 */
int 
NMAPReadResponseAndCount(Connection *conn, unsigned long *count)
{
    int ccode;
    char *newLine;

    do {
        newLine = FindEndOfLine(conn);
        if (newLine) {
            if (*newLine == '\n') {
                newLine++;
                ccode = atol(conn->receive.read);
                if (ccode != 6000) {
                    if ((ccode > 999) && (ccode < 10000) && ((conn->receive.read[4] == ' ') || (conn->receive.read[4] == '-'))) {
                        *count = atol(conn->receive.read + 5);
                        conn->receive.read = newLine;
                        return(ccode);
                    } 

                    return(NMAP_ERROR_INVALID_RESPONSE_CODE);
                }

                if (conn->client.cb) {
                    if (((NMAPOutOfBoundsCallback)conn->client.cb)(conn->client.data, conn->receive.read + 5, newLine)) {
                        conn->receive.read = newLine;
                        continue;
                    }

                    return(NMAP_ERROR_CALLBACK_FUNCTION_FAILED);
                }

                return(NMAP_ERROR_CALLBACK_FUNCTION_NOT_DEFINED);
            }

            return(NMAP_ERROR_LINE_TOO_LONG);
        }

        break;
    } while (TRUE);

    return(NMAP_ERROR_COMM);
}


/** NMAPRunCommandF
 *
 *  Equivalent to NMAPSendCommandF() followed by NMAPReadResponse().
 *
 *  Return Value
 *     - server response code
 *     - -1 upon connection error
 */

int 
NMAPRunCommandF(Connection *conn, char *response, size_t length, const char *format, ...)
{
    int i;
    va_list ap;
    Connection *c = conn;

    do {
        if (CONN_BUFSIZE < c->send.remaining) {
            va_start(ap, format);
            i = XplVsnprintf(c->send.write, CONN_BUFSIZE, format, ap);
            va_end(ap);

            if (i >= 0) {
                c->send.write += i;
                c->send.remaining -= i;

                CONN_TCP_FLUSH(c, c->send.read, c->send.write, i);
                if (i > 0) {
                    c->send.read = c->send.write = c->send.buffer;
                    c->send.remaining = CONN_TCP_MTU;

                    return NMAPReadResponse(conn, response, length, TRUE);
                }
            }
        }

        CONN_TCP_FLUSH(c, c->send.read, c->send.write, i);
        if (i > 0) {
            c->send.read = c->send.write = c->send.buffer;
            c->send.remaining = CONN_TCP_MTU;

            continue;
        }

        return(-1);
    } while (TRUE);
}






/** Read answer from NMAP connection. Optionally parses and returns the NMAP response number.
 *  \param[in] conn  The NMAP Connection struct.
 *  \param[out] response  The buffer to read the response into.
 *  \param[in] length  The maximum number of characters to read into the \c response buffer.
 *  \param[in] checkForResult  Enables or disables parsing out the response number.
 *  \return If the response number was requested and correctly parsed, then it returns
 *    that; -1 otherwise.
 */
int
NMAPReadAnswer(Connection *conn, unsigned char *response, size_t length, BOOL checkForResult)
{
    int len;
    int result = -1;
    size_t count;
    unsigned char *cur;

    len = ConnReadAnswer(conn, response, length);
    if (len > 0) {
        cur = response + 4;
        if (checkForResult && ((*cur == ' ') || (*cur == '-') || ((cur = strchr(response, ' ')) != NULL))) {
            *cur++ = '\0';

            result = atoi(response);

            count = len - (cur - response);
            memmove(response, cur, count);
            response[count] = '\0';
        } else {
            result = atoi(response);
        }
    }

    return(result);
}
                                                                                                                                                                            
int
NMAPReadAnswerLine(Connection *conn, unsigned char *response, size_t length, BOOL checkForResult)
{
    int len;
    int result = -1 ; 
    size_t count;
    unsigned char *cur;

    len = ConnReadLine(conn, response, length);
    if (len > 0) {
        cur = response + 4;
        if (checkForResult && ((*cur == ' ') || (*cur == '-') || ((cur = strchr(response, ' ')) != NULL))) {
            *cur++ = '\0';

	    result = atoi(response);
	    
            count = len - (cur - response);
            memmove(response, cur, count);
            response[count] = '\0';
        } else {
            result = atoi(response);
        }
    }

    return(result);
}

__inline static Connection *
NmapConnect(unsigned char *address, struct sockaddr_in *addr, int port, TraceDestination *destination)
{
    int ccode;
    Connection *conn;

    if (address || addr) {
        conn = ConnAlloc(TRUE);
    } else {
        return(NULL);
    }

    if (conn) {
        memset(&conn->socketAddress, 0, sizeof(struct sockaddr_in));

        if (address) {
            conn->socketAddress.sin_family = AF_INET;
            conn->socketAddress.sin_addr.s_addr = inet_addr(address);
            conn->socketAddress.sin_port = htons(port);
        } else {
            conn->socketAddress.sin_family = addr->sin_family;
            conn->socketAddress.sin_addr.s_addr = addr->sin_addr.s_addr;
            conn->socketAddress.sin_port = addr->sin_port;
        }

        conn->socket = IPsocket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
        if (conn->socket != -1) {
            ccode = IPconnect(conn->socket, (struct sockaddr *)&(conn->socketAddress), sizeof(struct sockaddr_in));
            CONN_TRACE_BEGIN(conn, CONN_TYPE_NMAP, destination);
            CONN_TRACE_EVENT(conn, CONN_TRACE_EVENT_CONNECT);
            if (!ccode) {
                /* Set TCP non blocking io */
                ccode = 1;
                setsockopt(conn->socket, IPPROTO_TCP, 1, (unsigned char *)&ccode, sizeof(ccode));

                return(conn);
            }

            CONN_TRACE_ERROR(conn, "connect", ccode);
            CONN_TRACE_EVENT(conn, CONN_TRACE_EVENT_CLOSE);
            CONN_TRACE_END(conn);
            IPclose(conn->socket);
            conn->socket = -1;
        }

        ConnFree(conn);
    }

    return(NULL);
}

Connection *
NMAPConnect(unsigned char *address, struct sockaddr_in *addr)
{
    return(NmapConnect(address, addr, NMAP_PORT, NULL));
}

Connection *
NMAPConnectEx(unsigned char *address, struct sockaddr_in *addr, TraceDestination *destination)
{
    return(NmapConnect(address, addr, NMAP_PORT, destination));
}
                                                                                                                                                                            
BOOL 
NMAPEncrypt(Connection *conn, unsigned char *response, int length, BOOL force)
{
    int ccode;
    BOOL result;
    Connection *c = conn;

    result = c->ssl.enable = FALSE;
    if (!force && XplIsLocalIPAddress(conn->socketAddress.sin_addr.s_addr)) {
        return(TRUE);
    }

    if (NMAPLibrary.context) {
        setsockopt(c->socket, IPPROTO_TCP, 1, (unsigned char *)&ccode, sizeof(ccode));

        if (((ccode = ConnWrite(c, "CAPA\r\n", 6)) != -1) 
                && ((ccode = ConnFlush(c)) != -1)) {
            while ((ccode = NMAPReadAnswer(c, response, length, TRUE)) == 2001) {
                if (strcmp(response, "TLS Encryption Extension") != 0) {
                    continue;
                }

                result = TRUE;
            }

            if (ccode != 1000) {
                result = FALSE;
            }
        }

        if (result 
                && ((c->ssl.context = SSL_new(NMAPLibrary.context)) != NULL) 
                && ((ccode = ConnWrite(c, "TLS\r\n", 5)) != -1) 
                && ((ccode = ConnFlush(c)) != -1) 
                && ((ccode = NMAPReadAnswer(c, response, length, TRUE)) == 1000) 
                && ((ccode = SSL_set_bsdfd(c->ssl.context, c->socket)) == 1) 
                && ((ccode = SSL_connect(c->ssl.context)) == 1)) {
            CONN_TRACE_EVENT(c, CONN_TRACE_EVENT_SSL_CONNECT);
            c->ssl.enable = TRUE;
            return(TRUE);
        }

        if (c->ssl.context) {
            SSL_free(c->ssl.context);
            c->ssl.context = NULL;
        }
    }

    return(FALSE);
}

BOOL 
NMAPAuthenticate(Connection *conn, unsigned char *response, int length)
{
    int ccode;

    ccode = NMAPReadAnswer(conn, response, length, TRUE);
    switch (ccode) {
        case 1000: {
            NMAPEncrypt(conn, response, length, FALSE);
            return(TRUE);
        }

        case 4242: {
            int                i;
            unsigned char    *ptr;
            unsigned char    *salt;
            unsigned char    digest[16];
            unsigned char    message[33];
            MD5_CTX        context;

            ptr = strchr(response, '<');
            if (ptr) {
                salt = ++ptr;

                if ((ptr = strchr(ptr, '>')) != NULL) {
                    *ptr = '\0';
                }

                MD5_Init(&context);

                MD5_Update(&context, salt, strlen(salt));
                MD5_Update(&context, NMAPLibrary.access, NMAP_HASH_SIZE);

                MD5_Final(digest, &context);

                NMAPEncrypt(conn, response, length, FALSE);

                ConnWrite(conn, "AUTH ", 5);

                for (i = 0; i < 16; i++) {
                    sprintf(message + (i * 2), "%02x", digest[i]);
                }

                ConnWrite(conn, message, 32);
                ConnWrite(conn, "\r\n", 2);

                if ((i = ConnFlush(conn)) == 39) {
                    if ((i = NMAPReadAnswer(conn, response, length, TRUE)) == 1000) {
                        return(TRUE);
                    }
                }
            }

            /*
                Fall through to the default case statement.
            */
        }

        default: {
            break;
        }
    }

    return(FALSE);
}

void 
NMAPQuit(Connection *conn)
{
    ConnWrite(conn, "QUIT\r\n", 6);

    CONN_TCP_CLOSE(conn);

    return;
}

RegistrationStates 
NMAPRegister(const unsigned char *dn, unsigned long queue, unsigned short port)
{
    int j;
    int ccode;
    unsigned long i;
    unsigned char response[CONN_BUFSIZE + 1];
    BOOL result;
    MDBValueStruct *servers = NULL;
    Connection *conn = NULL;

    if (dn) {
        NMAPLibrary.state = REGISTRATION_ALLOCATING;

        servers = MDBCreateValueStruct(NMAPLibrary.directoryHandle, MsgGetServerDN(NULL));
    } else {
        NMAPLibrary.state = REGISTRATION_FAILED;
        return(NMAPLibrary.state);
    }

    if (servers != NULL) {
        result = MsgReadIP(dn, MSGSRV_A_MONITORED_QUEUE, servers);
    } else {
        NMAPLibrary.state = REGISTRATION_FAILED;
        return(NMAPLibrary.state);
    }

    /* Couldn't find a configuration, try localhost */
    if (!result) {
        XplConsolePrintf("Couldn't find configuration object for %s, attempting to connect to NMAP on 127.0.0.1\n", dn);
        MDBAddValue("127.0.0.1", servers);
        result = TRUE;
    }

    if (result) {
        result = FALSE;
        NMAPLibrary.state = REGISTRATION_CONNECTING;

        for (i = 0; (i < servers->Used) && (NMAPLibrary.state == REGISTRATION_CONNECTING); i++) {
            NMAPLibrary.state = REGISTRATION_CONNECTING;

            do {
                conn = NmapConnect(servers->Value[i], NULL, HULA_QUEUE_PORT, NULL);
                if (conn) {
                    NMAPLibrary.state = REGISTRATION_REGISTERING;
                    break;
                }

                for (j = 0; (j < 15) && (NMAPLibrary.state == REGISTRATION_CONNECTING); j++) {
                    XplDelay(1000);
                }
            } while (NMAPLibrary.state == REGISTRATION_CONNECTING);

            if (NMAPLibrary.state == REGISTRATION_REGISTERING) {
                if (NMAPAuthenticate(conn, response, CONN_BUFSIZE)) {
                    ccode = ConnWriteF(conn, "QWAIT %lu %d %s%s%lu\r\n", queue, ntohs(port), MsgGetServerDN(NULL), dn, queue);
                    if (ccode > 0) {
                        ConnFlush(conn);

                        if (NMAPReadAnswer(conn, response, CONN_BUFSIZE, TRUE) == 1000) {
                            NMAPLibrary.state = REGISTRATION_COMPLETED;
                        }
                    }
                }

                NMAPQuit(conn);
            }

            if (conn) {
                ConnFree(conn);
            }
        }

        if (NMAPLibrary.state != REGISTRATION_COMPLETED) {
            NMAPLibrary.state = REGISTRATION_FAILED;
        }
    } else {
        NMAPLibrary.state = REGISTRATION_FAILED;
    }

    MDBDestroyValueStruct(servers);

    return(NMAPLibrary.state);
}

void 
NMAPSetEncryption(SSL_CTX *context)
{
    NMAPLibrary.context = context;

    return;
}

SSL_CTX *
NMAPSSLContextAlloc(void)
{
    ConnSSLConfiguration config;
    
    config.method = SSLv23_client_method;
    config.options = SSL_OP_DONT_INSERT_EMPTY_FRAGMENTS;
    config.mode = SSL_MODE_AUTO_RETRY;
    config.cipherList = NULL;
    config.certificate.type = SSL_FILETYPE_PEM;
    config.certificate.file = MsgGetTLSCertPath(NULL);
    config.key.type = SSL_FILETYPE_PEM;
    config.key.file = MsgGetTLSKeyPath(NULL);

    return ConnSSLContextAlloc(&config);
}

BOOL 
NMAPInitialize(MDBHandle directoryHandle)
{
    BOOL result = FALSE;
    MDBValueStruct *v;

    if (directoryHandle) {
        NMAPLibrary.directoryHandle = directoryHandle;

        v = MDBCreateValueStruct(NMAPLibrary.directoryHandle, NULL);
        if (v) {
            MDBRead(MSGSRV_ROOT, MSGSRV_A_ACL, v);
            if (v->Used) {
                result = HashCredential(MsgGetServerDN(NULL), v->Value[0], NMAPLibrary.access);
            }

            MDBDestroyValueStruct(v);
        }
    }

    return(result);
}
